#
# Copyright (c) 2016 Nordic Semiconductor ASA
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#   1. Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright notice, this
#   list of conditions and the following disclaimer in the documentation and/or
#   other materials provided with the distribution.
#
#   3. Neither the name of Nordic Semiconductor ASA nor the names of other
#   contributors to this software may be used to endorse or promote products
#   derived from this software without specific prior written permission.
#
#   4. This software must only be used in or with a processor manufactured by Nordic
#   Semiconductor ASA, or in or with a processor manufactured by a third party that
#   is used in combination with a processor manufactured by Nordic Semiconductor.
#
#   5. Any software provided in binary or object form under this license must not be
#   reverse engineered, decompiled, modified and/or disassembled.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import json
import unittest

from nordicsemi.dfu.manifest import ManifestGenerator, Manifest
from nordicsemi.dfu.model import HexType
from nordicsemi.dfu.package import FirmwareKeys


class TestManifest(unittest.TestCase):
    def setUp(self):
        self.firmwares_data_a = {}

        self.firmwares_data_a[HexType.APPLICATION] = {
            FirmwareKeys.BIN_FILENAME: "app_fw.bin",
            FirmwareKeys.DAT_FILENAME: "app_fw.dat",
            FirmwareKeys.ENCRYPT: False}

        self.firmwares_data_a[HexType.SD_BL] = {
            FirmwareKeys.BIN_FILENAME: "sd_bl_fw.bin",
            FirmwareKeys.DAT_FILENAME: "sd_bl_fw.dat",
            FirmwareKeys.BL_SIZE: 50,
            FirmwareKeys.SD_SIZE: 90
        }

        self.firmwares_data_b = {}

        self.firmwares_data_b[HexType.APPLICATION] = {
            FirmwareKeys.BIN_FILENAME: "app_fw.bin",
            FirmwareKeys.DAT_FILENAME: "app_fw.dat"
        }

        self.firmwares_data_b[HexType.BOOTLOADER] = {
            FirmwareKeys.BIN_FILENAME: "bootloader_fw.bin",
            FirmwareKeys.DAT_FILENAME: "bootloader_fw.dat"
        }

        self.firmwares_data_c = {}

        self.firmwares_data_c[HexType.SOFTDEVICE] = {
            FirmwareKeys.BIN_FILENAME: "softdevice_fw.bin",
            FirmwareKeys.DAT_FILENAME: "softdevice_fw.dat",
        }

    def test_generate_manifest(self):
        r = ManifestGenerator(self.firmwares_data_a)

        _json = json.loads(r.generate_manifest())

        # Test for presence of attributes in document
        self.assertIn('manifest', _json)

        manifest = _json['manifest']
        self.assertIn('application', manifest)

        application = manifest['application']
        self.assertIn('dat_file', application)
        self.assertIn('bin_file', application)

        # Test for values in document
        self.assertEqual("app_fw.bin", application['bin_file'])
        self.assertEqual("app_fw.dat", application['dat_file'])

        # Test softdevice_bootloader
        bl_sd = manifest['softdevice_bootloader']
        self.assertIsNotNone(bl_sd)
        self.assertEqual(90, bl_sd['info_read_only_metadata']['sd_size'])
        self.assertEqual(50, bl_sd['info_read_only_metadata']['bl_size'])

        # Test for values in document
        self.assertEqual("sd_bl_fw.bin", bl_sd['bin_file'])
        self.assertEqual("sd_bl_fw.dat", bl_sd['dat_file'])

    def test_manifest_a(self):
        r = ManifestGenerator(self.firmwares_data_a)
        m = Manifest.from_json(r.generate_manifest())
        self.assertIsNotNone(m)
        self.assertIsNotNone(m.application)
        self.assertEqual("app_fw.bin", m.application.bin_file)
        self.assertEqual("app_fw.dat", m.application.dat_file)
        self.assertIsNone(m.bootloader)
        self.assertIsNone(m.softdevice)
        self.assertIsNotNone(m.softdevice_bootloader)
        self.assertEqual(90, m.softdevice_bootloader.info_read_only_metadata.sd_size)
        self.assertEqual(50, m.softdevice_bootloader.info_read_only_metadata.bl_size)
        self.assertEqual("sd_bl_fw.bin", m.softdevice_bootloader.bin_file)
        self.assertEqual("sd_bl_fw.dat", m.softdevice_bootloader.dat_file)

    def test_manifest_b(self):
        r = ManifestGenerator(self.firmwares_data_b)
        m = Manifest.from_json(r.generate_manifest())
        self.assertIsNotNone(m)
        self.assertIsNotNone(m.application)
        self.assertEqual("app_fw.bin", m.application.bin_file)
        self.assertEqual("app_fw.dat", m.application.dat_file)
        self.assertIsNotNone(m.bootloader)
        self.assertEqual("bootloader_fw.bin", m.bootloader.bin_file)
        self.assertEqual("bootloader_fw.dat", m.bootloader.dat_file)
        self.assertIsNone(m.softdevice)
        self.assertIsNone(m.softdevice_bootloader)


    def test_manifest_c(self):
        r = ManifestGenerator(self.firmwares_data_c)
        m = Manifest.from_json(r.generate_manifest())
        self.assertIsNotNone(m)
        self.assertIsNone(m.application)
        self.assertIsNone(m.bootloader)
        self.assertIsNotNone(m.softdevice)
        self.assertEqual('softdevice_fw.bin', m.softdevice.bin_file)
        self.assertEqual('softdevice_fw.dat', m.softdevice.dat_file)
        self.assertIsNone(m.softdevice_bootloader)

if __name__ == '__main__':
    unittest.main()



#
# Copyright (c) 2016 Nordic Semiconductor ASA
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#   1. Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright notice, this
#   list of conditions and the following disclaimer in the documentation and/or
#   other materials provided with the distribution.
#
#   3. Neither the name of Nordic Semiconductor ASA nor the names of other
#   contributors to this software may be used to endorse or promote products
#   derived from this software without specific prior written permission.
#
#   4. This software must only be used in or with a processor manufactured by Nordic
#   Semiconductor ASA, or in or with a processor manufactured by a third party that
#   is used in combination with a processor manufactured by Nordic Semiconductor.
#
#   5. Any software provided in binary or object form under this license must not be
#   reverse engineered, decompiled, modified and/or disassembled.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
import unittest
from nordicsemi.dfu.init_packet_pb import InitPacketPB, DFUType, SigningTypes, HashTypes
import nordicsemi.dfu.dfu_cc_pb2 as pb

HASH_BYTES_A = b'123123123123'
HASH_BYTES_B = b'434343434343'
HASH_TYPE = HashTypes.SHA256
DFU_TYPE = DFUType.APPLICATION
SIGNATURE_BYTES_A = b'234827364872634876234'
SIGNATURE_TYPE = SigningTypes.ECDSA_P256_SHA256
SD_REQ_A = [1, 2, 3, 4]
FIRMWARE_VERSION_A = 0xaaaa
HARDWARE_VERSION_A = 0xbbbb
ILLEGAL_VERSION = 0xaaaaaaaa1
SD_SIZE = 0x11
APP_SIZE = 0x233
BL_SIZE = 0x324


class TestPackage(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_init_command(self):
        init_command_serialized = InitPacketPB(hash_bytes=HASH_BYTES_B, hash_type=HASH_TYPE,
                                               dfu_type=DFU_TYPE, sd_req=SD_REQ_A, fw_version=FIRMWARE_VERSION_A,
                                               hw_version=HARDWARE_VERSION_A, sd_size=SD_SIZE, app_size=APP_SIZE,
                                               bl_size=BL_SIZE).get_init_command_bytes()

        init_command = pb.InitCommand()
        init_command.ParseFromString(init_command_serialized)

        self.assertEqual(init_command.hash.hash, HASH_BYTES_B)
        self.assertEqual(init_command.hash.hash_type, pb.SHA256)
        self.assertEqual(init_command.type, pb.APPLICATION)
        self.assertEqual(init_command.fw_version, FIRMWARE_VERSION_A)
        self.assertEqual(init_command.hw_version, HARDWARE_VERSION_A)
        self.assertEqual(init_command.app_size, APP_SIZE)
        self.assertEqual(init_command.sd_size, SD_SIZE)
        self.assertEqual(init_command.bl_size, BL_SIZE)
        self.assertEqual(init_command.sd_req, SD_REQ_A)

    def test_init_command_wrong_size(self):
        def test_size(dfu_type, sd_size, app_size, bl_size, expect_failed):
            failed = False
            try:
                InitPacketPB(hash_bytes=HASH_BYTES_B, hash_type=HASH_TYPE,
                             dfu_type=dfu_type,
                             sd_size=sd_size,
                             app_size=app_size,
                             bl_size=bl_size)
            except RuntimeError:
                failed = True

            self.assertEqual(failed, expect_failed)

        test_size(DFUType.APPLICATION, sd_size=SD_SIZE, app_size=0, bl_size=BL_SIZE, expect_failed=True)
        test_size(DFUType.APPLICATION, sd_size=SD_SIZE, app_size=APP_SIZE, bl_size=BL_SIZE, expect_failed=False)
        test_size(DFUType.BOOTLOADER, sd_size=SD_SIZE, app_size=APP_SIZE, bl_size=0, expect_failed=True)
        test_size(DFUType.BOOTLOADER, sd_size=SD_SIZE, app_size=APP_SIZE, bl_size=BL_SIZE, expect_failed=False)
        test_size(DFUType.SOFTDEVICE, sd_size=0, app_size=APP_SIZE, bl_size=BL_SIZE, expect_failed=True)
        test_size(DFUType.SOFTDEVICE, sd_size=SD_SIZE, app_size=APP_SIZE, bl_size=BL_SIZE, expect_failed=False)
        test_size(DFUType.SOFTDEVICE_BOOTLOADER, sd_size=0, app_size=APP_SIZE, bl_size=BL_SIZE, expect_failed=True)
        test_size(DFUType.SOFTDEVICE_BOOTLOADER, sd_size=SD_SIZE, app_size=APP_SIZE, bl_size=0, expect_failed=True)
        test_size(DFUType.SOFTDEVICE_BOOTLOADER, sd_size=SD_SIZE, app_size=APP_SIZE, bl_size=BL_SIZE,
                  expect_failed=False)

    def test_init_packet(self):
        failed = False
        init_packet = InitPacketPB(hash_bytes=HASH_BYTES_A, hash_type=HASH_TYPE, dfu_type=DFU_TYPE, app_size=APP_SIZE)

        init_packet.set_signature(signature=SIGNATURE_BYTES_A, signature_type=SIGNATURE_TYPE)
        init_packet_serialized = init_packet.get_init_packet_pb_bytes()

        init_packet = pb.Packet()
        init_packet.ParseFromString(init_packet_serialized)
        self.assertEqual(init_packet.signed_command.command.init.hash.hash, HASH_BYTES_A)
        self.assertEqual(init_packet.signed_command.command.op_code, pb.INIT)


if __name__ == '__main__':
    unittest.main()



#
# Copyright (c) 2019 Nordic Semiconductor ASA
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#   1. Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright notice, this
#   list of conditions and the following disclaimer in the documentation and/or
#   other materials provided with the distribution.
#
#   3. Neither the name of Nordic Semiconductor ASA nor the names of other
#   contributors to this software may be used to endorse or promote products
#   derived from this software without specific prior written permission.
#
#   4. This software must only be used in or with a processor manufactured by Nordic
#   Semiconductor ASA, or in or with a processor manufactured by a third party that
#   is used in combination with a processor manufactured by Nordic Semiconductor.
#
#   5. Any software provided in binary or object form under this license must not be
#   reverse engineered, decompiled, modified and/or disassembled.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import sys
from nordicsemi.lister.windows.lister_win32 import Win32Lister
from nordicsemi.lister.unix.unix_lister import UnixLister


class DeviceLister:
    def __init__(self):
        if sys.platform == 'win32':
            self.lister_backend = Win32Lister()
        elif 'linux' in sys.platform:
            self.lister_backend = UnixLister()
        elif sys.platform == 'darwin':
            self.lister_backend = UnixLister()
        else:
            self.lister_backend = None

    def enumerate(self):
        if self.lister_backend:
            return self.lister_backend.enumerate()
        return []

    def get_device(self, get_all=False, **kwargs):
        devices = self.enumerate()
        matching_devices = []
        for dev in devices:
            if "vendor_id" in kwargs and kwargs["vendor_id"].lower() != dev.vendor_id.lower():
                continue
            if "product_id" in kwargs and kwargs["product_id"].lower() != dev.product_id.lower():
                continue
            if "serial_number" in kwargs and (kwargs["serial_number"].lower().lstrip('0') !=
                                              dev.serial_number.lower().lstrip('0')):
                continue
            if "com" in kwargs and not dev.has_com_port(kwargs["com"]):
                continue

            matching_devices.append(dev)

        if not get_all:
            if len(matching_devices) == 0:
                return
            return matching_devices[0]
        return matching_devices



#
# Copyright (c) 2016 Nordic Semiconductor ASA
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#   1. Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright notice, this
#   list of conditions and the following disclaimer in the documentation and/or
#   other materials provided with the distribution.
#
#   3. Neither the name of Nordic Semiconductor ASA nor the names of other
#   contributors to this software may be used to endorse or promote products
#   derived from this software without specific prior written permission.
#
#   4. This software must only be used in or with a processor manufactured by Nordic
#   Semiconductor ASA, or in or with a processor manufactured by a third party that
#   is used in combination with a processor manufactured by Nordic Semiconductor.
#
#   5. Any software provided in binary or object form under this license must not be
#   reverse engineered, decompiled, modified and/or disassembled.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
import ipaddress
import signal

"""nrfutil command line tool."""
import os
import sys
import click
import time
import logging
import re
sys.path.append(os.getcwd())

from nordicsemi.dfu.bl_dfu_sett import BLDFUSettings
from nordicsemi.dfu.dfu import Dfu
from nordicsemi.dfu.dfu_transport import DfuEvent, TRANSPORT_LOGGING_LEVEL
from nordicsemi.dfu.dfu_transport_serial import DfuTransportSerial
from nordicsemi.dfu.package import Package
from nordicsemi import version as nrfutil_version
from nordicsemi.dfu.signing import Signing
from nordicsemi.zigbee.prod_config import ProductionConfig, ProductionConfigWrongException, ProductionConfigTooLargeException
from pc_ble_driver_py.exceptions import NordicSemiException
from nordicsemi.lister.device_lister import DeviceLister
import spinel.util as util

logger = logging.getLogger(__name__)

def ble_driver_init(conn_ic_id):
    global BLEDriver, Flasher, DfuTransportBle, config
    from pc_ble_driver_py import config
    config.__conn_ic_id__ = conn_ic_id
    from pc_ble_driver_py.ble_driver    import BLEDriver, Flasher
    from nordicsemi.dfu.dfu_transport_ble import DfuTransportBle

def display_sec_warning():
    default_key_warning = """
|===============================================================|
|##      ##    ###    ########  ##    ## #### ##    ##  ######  |
|##  ##  ##   ## ##   ##     ## ###   ##  ##  ###   ## ##    ## |
|##  ##  ##  ##   ##  ##     ## ####  ##  ##  ####  ## ##       |
|##  ##  ## ##     ## ########  ## ## ##  ##  ## ## ## ##   ####|
|##  ##  ## ######### ##   ##   ##  ####  ##  ##  #### ##    ## |
|##  ##  ## ##     ## ##    ##  ##   ###  ##  ##   ### ##    ## |
| ###  ###  ##     ## ##     ## ##    ## #### ##    ##  ######  |
|===============================================================|
|The security key you provided is insecure, as it part of a     |
|known set of keys that have been widely distributed. Do NOT use|
|it in your final product or your DFU procedure may be          |
|compromised and at risk of malicious attacks.                  |
|===============================================================|
"""
    click.echo("{}".format(default_key_warning))

def display_nokey_warning():
    default_nokey_warning = """
|===============================================================|
|##      ##    ###    ########  ##    ## #### ##    ##  ######  |
|##  ##  ##   ## ##   ##     ## ###   ##  ##  ###   ## ##    ## |
|##  ##  ##  ##   ##  ##     ## ####  ##  ##  ####  ## ##       |
|##  ##  ## ##     ## ########  ## ## ##  ##  ## ## ## ##   ####|
|##  ##  ## ######### ##   ##   ##  ####  ##  ##  #### ##    ## |
|##  ##  ## ##     ## ##    ##  ##   ###  ##  ##   ### ##    ## |
| ###  ###  ##     ## ##     ## ##    ## #### ##    ##  ######  |
|===============================================================|
|You are not providing a signature key, which means the DFU     |
|files will not be signed, and are vulnerable to tampering.     |
|This is only compatible with a signature-less bootloader and is|
|not suitable for production environments.                      |
|===============================================================|
"""
    click.echo("{}".format(default_nokey_warning))

def display_debug_warning():
    debug_warning = """
|===============================================================|
|##      ##    ###    ########  ##    ## #### ##    ##  ######  |
|##  ##  ##   ## ##   ##     ## ###   ##  ##  ###   ## ##    ## |
|##  ##  ##  ##   ##  ##     ## ####  ##  ##  ####  ## ##       |
|##  ##  ## ##     ## ########  ## ## ##  ##  ## ## ## ##   ####|
|##  ##  ## ######### ##   ##   ##  ####  ##  ##  #### ##    ## |
|##  ##  ## ##     ## ##    ##  ##   ###  ##  ##   ### ##    ## |
| ###  ###  ##     ## ##     ## ##    ## #### ##    ##  ######  |
|===============================================================|
|You are generating a package with the debug bit enabled in the |
|init packet. This is only compatible with a debug bootloader   |
|and is not suitable for production.                            |
|===============================================================|
"""
    click.echo("{}".format(debug_warning))

def display_settings_backup_warning():
    debug_warning = """
Note: Generating a DFU settings page with backup page included.
This is only required for bootloaders from nRF5 SDK 15.1 and newer.
If you want to skip backup page generation, use --no-backup option."""
    click.echo("{}".format(debug_warning))

def int_as_text_to_int(value):
    try:
        if value[:2].lower() == '0x':
            return int(value[2:], 16)
        elif value[:1] == '0':
            return int(value, 8)
        return int(value, 10)
    except ValueError:
        raise NordicSemiException('%s is not a valid integer' % value)

# TODO: Create query function that maps query-result strings with functions
def query_func(question, default=False):
    """
    Ask a string question
    No input defaults to "no" which results in False
    """
    valid = {"yes": True, "y": True, "no": False, "n": False}
    if default is True:
        prompt = " [Y/n]"
    else:
        prompt = " [y/N]"

    while True:
        print("%s %s" % (question, prompt))
        choice = input().lower()
        if choice == '':
            return default
        elif choice in valid:
            return valid[choice]
        else:
            print("Please respond with y/n")

def pause():
    while True:
        try:
            input()
        except (KeyboardInterrupt, EOFError):
            break

class BasedIntOrNoneParamType(click.ParamType):
    name = 'Integer'

    def convert(self, value, param, ctx):
        try:
            if value.lower() == 'none':
                return 'none'
            return int_as_text_to_int(value)
        except NordicSemiException:
            self.fail('%s is not a valid integer' % value, param, ctx)

BASED_INT_OR_NONE = BasedIntOrNoneParamType()

class BasedIntParamType(BasedIntOrNoneParamType):
    name = 'Integer'

BASED_INT = BasedIntParamType()

class TextOrNoneParamType(click.ParamType):
    name = 'Text'

    def convert(self, value, param, ctx):
        return value

TEXT_OR_NONE = TextOrNoneParamType()

BOOT_VALIDATION_ARGS = [
    'NO_VALIDATION',
    'VALIDATE_GENERATED_CRC',
    'VALIDATE_GENERATED_SHA256',
    'VALIDATE_ECDSA_P256_SHA256',
]
DEFAULT_BOOT_VALIDATION = 'VALIDATE_GENERATED_CRC'

KEY_CHOICE = ['pk', 'sk']
KEY_FORMAT = [
    'hex',
    'code',
    'pem',
    'dbgcode',
]


class OptionRequiredIf(click.Option):

    def full_process_value(self, ctx, value):
        value = super().full_process_value(ctx, value)
        if ('serial_number' not in ctx.params or not ctx.params['serial_number']) and value is None:
            msg = 'Required if "-snr" / "--serial-number" is not defined.'
            raise click.MissingParameter(ctx=ctx, param=self, message=msg)
        return value

@click.group()
@click.option('-v', '--verbose',
              help='Increase verbosity of output. Can be specified more than once (up to -v -v -v -v).',
              count=True)
@click.option('-o', '--output',
              help='Log output to file',
              metavar='<filename>')
def cli(verbose, output):
    #click.echo('verbosity: %s' % verbose)
    if verbose == 0:
        log_level = logging.ERROR
    elif verbose == 1:
        log_level = logging.WARNING
    elif verbose == 2:
        log_level = logging.INFO
    elif verbose == 3:
        log_level = logging.DEBUG
    else:
        # Custom level, logs all the bytes sent/received over the wire/air
        log_level = TRANSPORT_LOGGING_LEVEL

    logging.basicConfig(format='%(asctime)s %(message)s', level=log_level)

    if (output):
        root = logging.getLogger('')
        fh = logging.FileHandler(output)
        fh.setLevel(log_level)
        fh.setFormatter(logging.Formatter('%(asctime)s %(message)s'))
        root.addHandler(fh)

@cli.command()
def version():
    """Display nrfutil version."""
    click.echo("nrfutil version {}".format(nrfutil_version.NRFUTIL_VERSION))
    logger.info("PyPi URL: https://pypi.python.org/pypi/nrfutil")
    logger.debug("GitHub URL: https://github.com/NordicSemiconductor/pc-nrfutil")

@cli.group(short_help='Generate and display Bootloader DFU settings.')
def settings():
    """
    This set of commands supports creating and displaying bootloader settings.
    """
    pass

@settings.command(short_help='Generate a .hex file with Bootloader DFU settings.')
@click.argument('hex_file', required=True, type=click.Path())
@click.option('--family',
              help='nRF IC family: NRF51 or NRF52 or NRF52QFAB or NRF52810 or NRF52840',
              type=click.Choice(['NRF51', 'NRF52', 'NRF52QFAB', 'NRF52810', 'NRF52840']),
              required=True)
@click.option('--application',
              help='The application firmware file. This can be omitted if'
                    'the target IC does not contain an application in flash.'
                    'Requires --application-version or --application-version-string.',
              type=click.STRING)
@click.option('--application-version',
              help='The application version.',
              type=BASED_INT_OR_NONE)
@click.option('--application-version-string',
              help='The application version string, e.g. "2.7.31". Will be converted to an integer, e.g. 20731.',
              type=click.STRING)
@click.option('--bootloader-version',
              help='The bootloader version.',
              type=BASED_INT_OR_NONE,
              required=True)
@click.option('--bl-settings-version',
              help='The Bootloader settings version.'
              'Defined in nrf_dfu_types.h, the following apply to released SDKs:'
              '\n|SDK12.0.0 - SDK15.2.0|1|'
              '\n|SDK15.3.0 -          |2|',
              type=BASED_INT_OR_NONE,
              required=True)
@click.option('--start-address',
              help='Custom start address for the settings page. If not specified, '
                   'then the last page of the flash is used.',
              type=BASED_INT_OR_NONE)
@click.option('--no-backup',
              help='Do not overwrite DFU settings backup page. If not specified, '
                   'than the resulting .hex file will contain a copy of DFU settings, '
                   'that will overwrite contents of DFU settings backup page.',
              type=click.BOOL,
              is_flag=True,
              required=False)
@click.option('--backup-address',
              help='Address of the DFU settings backup page inside flash. '
                   'By default, the backup page address is placed one page below DFU settings. '
                   'The value is precalculated based on configured settings address '
                   '(<DFU_settings_address> - 0x1000).',
              type=BASED_INT_OR_NONE)
@click.option('--app-boot-validation',
              help='The method of boot validation for application.',
              required=False,
              type=click.Choice(BOOT_VALIDATION_ARGS))
@click.option('--sd-boot-validation',
              help='The method of boot validation for SoftDevice.',
              required=False,
              type=click.Choice(BOOT_VALIDATION_ARGS))
@click.option('--softdevice',
              help='The SoftDevice firmware file. Must be given if SD Boot Validation is used.',
              required=False,
              type=click.Path(exists=True, resolve_path=True, file_okay=True, dir_okay=False))
@click.option('--key-file',
              help='The private (signing) key in PEM format. Needed for ECDSA Boot Validation.',
              required=False,
              type=click.Path(exists=True, resolve_path=True, file_okay=True, dir_okay=False))
def generate(hex_file,
             family,
             application,
             application_version,
             application_version_string,
             bootloader_version,
             bl_settings_version,
             start_address,
             no_backup,
             backup_address,
             app_boot_validation,
             sd_boot_validation,
             softdevice,
             key_file):

    # The user can specify the application version with two different
    # formats. As an integer, e.g. 102130, or as a string
    # "10.21.30". Internally we convert to integer.
    if application_version_string:
        application_version_internal = convert_version_string_to_int(application_version_string)
        if application_version:
            click.echo('Warning: When both application-version-string and application-version are provided, only the string will be used.')
    else:
        application_version_internal = application_version

    if application is not None:
        if not os.path.isfile(application):
            raise click.FileError(application, hint="Application file not found")
        if application_version_internal is None:
            raise click.UsageError('--application-version or --application-version-string'
                                   ' required with application image.')

    if (no_backup is not None) and (backup_address is not None):
        raise click.BadParameter("Bootloader DFU settings backup page cannot be specified if backup is disabled.", param_hint='backup_address')

    if no_backup is None:
        no_backup = False

    if no_backup is False:
        display_settings_backup_warning()

    if (start_address is not None) and (backup_address is None):
        click.echo("WARNING: Using default offset in order to calculate bootloader settings backup page")

    if bl_settings_version == 1 and (app_boot_validation or sd_boot_validation):
        raise click.BadParameter("Bootloader settings version 1 does not support boot validation.", param_hint='bl_settings_version')

    # load signing key (if needed) only once
    if 'VALIDATE_ECDSA_P256_SHA256' in (app_boot_validation, sd_boot_validation):
        if not os.path.isfile(key_file):
            raise click.UsageError("Key file must be given when 'VALIDATE_ECDSA_P256_SHA256' boot validation is used")
        signer = Signing()
        default_key = signer.load_key(key_file)
        if default_key:
            display_sec_warning()
    else:
        signer = None

    if app_boot_validation and not application:
        raise click.UsageError("--application hex file must be set when using --app_boot_validation")

    if sd_boot_validation and not softdevice:
        raise click.UsageError("--softdevice hex file must be set when using --sd_boot_validation")

    # Default boot validation cases
    if app_boot_validation is None and application is not None and bl_settings_version == 2:
        app_boot_validation = DEFAULT_BOOT_VALIDATION
    if sd_boot_validation is None and softdevice is not None and bl_settings_version == 2:
        sd_boot_validation = DEFAULT_BOOT_VALIDATION

    sett = BLDFUSettings()
    sett.generate(arch=family, app_file=application, app_ver=application_version_internal, bl_ver=bootloader_version,
                  bl_sett_ver=bl_settings_version, custom_bl_sett_addr=start_address, no_backup=no_backup,
                  backup_address=backup_address, app_boot_validation_type=app_boot_validation,
                  sd_boot_validation_type=sd_boot_validation, sd_file=softdevice, signer=signer)
    sett.tohexfile(hex_file)

    click.echo("\nGenerated Bootloader DFU settings .hex file and stored it in: {}".format(hex_file))

    click.echo("{0}".format(str(sett)))

@settings.command(short_help='Display the contents of a .hex file with Bootloader DFU settings.')
@click.argument('hex_file', required=True, type=click.Path())

def display(hex_file):

    sett = BLDFUSettings()
    try:
        sett.fromhexfile(hex_file)
    except NordicSemiException as err:
        raise click.UsageError(err)

    click.echo("{0}".format(str(sett)))


@cli.group(short_help='Generate and display private and public keys.')
def keys():
    """
    This set of commands supports creating and displaying a private (signing) key
    as well as displaying the public (verification) key derived from a private key.
    Private keys are stored in PEM format.
    """
    pass

@keys.command(short_help='Generate a private key and store it in a file in PEM format.')
@click.argument('key_file', required=True, type=click.Path())

def generate(key_file):
    signer = Signing()

    if os.path.exists(key_file):
        if not query_func("File found at %s. Do you want to overwrite the file?" % key_file):
            click.echo('Key generation aborted.')
            return

    signer.gen_key(key_file)
    click.echo("Generated private key and stored it in: %s" % key_file)

@keys.command(short_help='Display the private key that is stored in a file in PEM format or a public key derived from it.')
@click.argument('key_file', required=True, type=click.Path())
@click.option('--key',
              help='(pk|sk) Display the public key (pk) or the private key (sk).',
              type=click.Choice(KEY_CHOICE),
              required=True)
@click.option('--format',
              help='(hex|code|pem) Display the key in hexadecimal format (hex), C code (code), or PEM (pem) format.',
              type=click.Choice(KEY_FORMAT),
              required=True)
@click.option('--out_file',
              help='If provided, save the output in file out_file.',
              type=click.STRING)

def display(key_file, key, format, out_file):
    signer = Signing()

    if not os.path.isfile(key_file):
        raise NordicSemiException("File not found: %s" % key_file)

    default_key = signer.load_key(key_file)
    if default_key:
        display_sec_warning()

    if format == "dbgcode":
        format = "code"
        dbg = True
    else:
        dbg = False

    if format == "code" and key == "sk":
        raise click.UsageError("Displaying the private key as code is not available.")

    if key == "pk":
        kstr = signer.get_vk(format, dbg)
    elif key == "sk":
        kstr = "\nWARNING: Security risk! Do not share the private key.\n\n"
        kstr = kstr + signer.get_sk(format, dbg)

    if not out_file:
        click.echo(kstr)
    else:
        with open(out_file, "w") as kfile:
            kfile.write(kstr)


@cli.group(short_help='Display or generate a DFU package (zip file).')
def pkg():
    """
    This set of commands supports Nordic DFU package generation.
    """
    pass


@pkg.command(short_help='Generate a zip file for performing DFU.')
@click.argument('zipfile',
                required=True,
                type=click.Path())
@click.option('--debug-mode',
              help='Debug mode switch, enables version check skipping.',
              type=click.BOOL,
              default=False,
              is_flag=True)
@click.option('--application',
              help='The application firmware file.',
              type=click.STRING)
@click.option('--application-version',
              help='The application version.',
              type=BASED_INT_OR_NONE)
@click.option('--application-version-string',
              help='The application version string, e.g. "2.7.31". Will be converted to an integer, e.g. 20731.',
              type=click.STRING)
@click.option('--bootloader',
              help='The bootloader firmware file.',
              type=click.STRING)
@click.option('--bootloader-version',
              help='The bootloader version.',
              type=BASED_INT_OR_NONE)
@click.option('--hw-version',
              help='The hardware version.',
              required=False,
              type=BASED_INT)
@click.option('--sd-req',
              help='The SoftDevice requirements. A comma-separated list of SoftDevice firmware IDs '
                   '(1 or more) of which one must be present on the target device. Each item on the '
                   'list must be a two- or four-digit hex number prefixed with \"0x\" (e.g. \"0x12\", '
                   '\"0x1234\").\n'
                   'A non-exhaustive list of well-known values to use with this option follows:'
                   '\n|s112_nrf52_6.0.0|0xA7|'
                   '\n|s112_nrf52_6.1.0|0xB0|'
                   '\n|s112_nrf52_6.1.1|0xB8|'
                   '\n|s112_nrf52_7.0.0|0xC4|'
                   '\n|s112_nrf52_7.0.1|0xCD|'
                   '\n|s112_nrf52_7.2.0|0x103|'
                   '\n|s112_nrf52_7.3.0|0x126|'
                   '\n|s113_nrf52_7.0.0|0xC3|'
                   '\n|s113_nrf52_7.0.1|0xCC|'
                   '\n|s113_nrf52_7.2.0|0x102|'
                   '\n|s113_nrf52_7.3.0|0x125|'
                   '\n|s122_nrf52_8.0.0|0xEA|'
                   '\n|s122_nrf52_8.1.1|0x112|'
                   '\n|s130_nrf51_1.0.0|0x67|'
                   '\n|s130_nrf51_2.0.0|0x80|'
                   '\n|s130_nrf51_2.0.1|0x87|'
                   '\n|s132_nrf52_2.0.0|0x81|'
                   '\n|s132_nrf52_2.0.1|0x88|'
                   '\n|s132_nrf52_3.0.0|0x8C|'
                   '\n|s132_nrf52_3.1.0|0x91|'
                   '\n|s132_nrf52_4.0.0|0x95|'
                   '\n|s132_nrf52_4.0.2|0x98|'
                   '\n|s132_nrf52_4.0.3|0x99|'
                   '\n|s132_nrf52_4.0.4|0x9E|'
                   '\n|s132_nrf52_4.0.5|0x9F|'
                   '\n|s132_nrf52_5.0.0|0x9D|'
                   '\n|s132_nrf52_5.1.0|0xA5|'
                   '\n|s132_nrf52_6.0.0|0xA8|'
                   '\n|s132_nrf52_6.1.0|0xAF|'
                   '\n|s132_nrf52_6.1.1|0xB7|'
                   '\n|s132_nrf52_7.0.0|0xC2|'
                   '\n|s132_nrf52_7.0.1|0xCB|'
                   '\n|s132_nrf52_7.2.0|0x101|'
                   '\n|s132_nrf52_7.3.0|0x124|'
                   '\n|s140_nrf52_6.0.0|0xA9|'
                   '\n|s140_nrf52_6.1.0|0xAE|'
                   '\n|s140_nrf52_6.1.1|0xB6|'
                   '\n|s140_nrf52_7.0.0|0xC1|'
                   '\n|s140_nrf52_7.0.1|0xCA|'
                   '\n|s140_nrf52_7.2.0|0x100|'
                   '\n|s140_nrf52_7.3.0|0x123|'
                   '\n|s212_nrf52_6.1.1|0xBC|'
                   '\n|s332_nrf52_6.1.1|0xBA|'
                   '\n|s340_nrf52_6.1.1|0xB9|',
              type=click.STRING,
              required=False,
              multiple=True)
@click.option('--sd-id',
              help='The new SoftDevice ID to be used as --sd-req for the Application update in case the ZIP '
                   'contains a SoftDevice and an Application.',
              type=click.STRING,
              multiple=True)
@click.option('--softdevice',
              help='The SoftDevice firmware file.',
              type=click.STRING)
@click.option('--sd-boot-validation',
              help='The method of boot validation for Softdevice.',
              required=False,
              type=click.Choice(BOOT_VALIDATION_ARGS))
@click.option('--app-boot-validation',
              help='The method of boot validation for application.',
              required=False,
              type=click.Choice(BOOT_VALIDATION_ARGS))
@click.option('--key-file',
              help='The private (signing) key in PEM format.',
              required=False,
              type=click.Path(exists=True, resolve_path=True, file_okay=True, dir_okay=False))
@click.option('--external-app',
              help='Indicates that the FW upgrade is intended to be passed through '
                   '(not applied on the receiving device)',
              type=click.BOOL, is_flag=True, default=False)
@click.option('--zigbee',
              help='Create an image and distribution package for Zigbee DFU server.',
              required=False,
              type=click.BOOL)
@click.option('--zigbee-manufacturer-id',
              help='Manufacturer ID to be used in Zigbee OTA header.',
              required=False,
              type=BASED_INT)
@click.option('--zigbee-image-type',
              help='Image type to be used in Zigbee OTA header.',
              required=False,
              type=BASED_INT)
@click.option('--zigbee-comment',
              help='Firmware comment to be used in Zigbee OTA header.',
              required=False,
              type=click.STRING)
@click.option('--zigbee-ota-hw-version',
              help='The zigbee OTA hw version.',
              required=False,
              type=BASED_INT_OR_NONE)
@click.option('--zigbee-ota-fw-version',
              help='The zigbee OTA fw version.',
              required=False,
              type=BASED_INT_OR_NONE)
@click.option('--zigbee-ota-min-hw-version',
              help='The zigbee OTA minimum hw version of Zigbee OTA Client.',
              required=False,
              type=BASED_INT_OR_NONE)
@click.option('--zigbee-ota-max-hw-version',
              help='The zigbee OTA maximum hw version of Zigbee OTA Client.',
              required=False,
              type=BASED_INT_OR_NONE)
def generate(zipfile,
           debug_mode,
           application,
           application_version,
           application_version_string,
           bootloader,
           bootloader_version,
           hw_version,
           sd_req,
           sd_id,
           softdevice,
           sd_boot_validation,
           app_boot_validation,
           key_file,
           external_app,
           zigbee,
           zigbee_manufacturer_id,
           zigbee_image_type,
           zigbee_comment,
           zigbee_ota_hw_version,
           zigbee_ota_fw_version,
           zigbee_ota_min_hw_version,
           zigbee_ota_max_hw_version):
    """
    Generate a zip package for distribution to apps that support Nordic DFU OTA.
    The application, bootloader, and SoftDevice files are converted to .bin if supplied as .hex files.
    For more information on the generated package, see:
    http://developer.nordicsemi.com/nRF5_SDK/doc/

    The following combinations are supported by this command:

    * BL only: Supported.

    * SD only: Supported (SD of same Major Version).

    * APP only: Supported (external or internal).

    * BL + SD: Supported.

    * BL + APP: Not supported (use two packages instead).

    * BL + SD + APP: Supported.

    * SD + APP: Supported (SD of same Major Version).
    """
    zipfile_path = zipfile

    # Check combinations
    if bootloader is not None and application is not None and softdevice is None:
        raise click.UsageError("Invalid combination: use two .zip packages instead.")

    if debug_mode is None:
        debug_mode = False

    # The user can specify the application version with two different
    # formats. As an integer, e.g. 102130, or as a string
    # "10.21.30". Internally we convert to integer.
    if application_version_string:
        application_version_internal = convert_version_string_to_int(application_version_string)
        if application_version:
            click.echo('Warning: When both application-version-string and application-version are provided, only the string will be used.')
    else:
        application_version_internal = application_version

    if application_version_internal == 'none':
        application_version_internal = None

    if bootloader_version == 'none':
        bootloader_version = None

    if hw_version == 'none':
        hw_version = None

    if external_app is None:
        external_app = False

    if zigbee_ota_hw_version == 'none':
        zigbee_ota_hw_version = None

    if zigbee_ota_fw_version == 'none':
        zigbee_ota_fw_version = None

    # Convert multiple value into a single instance
    if len(sd_req) > 1:
        raise click.BadParameter("Please specify SoftDevice requirements as a comma-separated list: --sd-req 0xXXXX,0xYYYY,...", param_hint='sd_req')
    elif len(sd_req) == 0:
        sd_req = None
    else:
        sd_req = sd_req[0]
        if sd_req == 'none':
            sd_req = None

    if len(sd_id) > 1:
        raise click.BadParameter("Please specify SoftDevice requirements as a comma-separated list: --sd-id 0xXXXX,0xYYYY,...", param_hint='sd_req')
    elif len(sd_id) == 0:
        sd_id = None
    else:
        sd_id = sd_id[0]
        if sd_id == 'none':
            sd_id = None

    # Initial consistency checks
    if application_version_internal is not None and application is None:
        raise click.UsageError("Application version with no image.")

    if bootloader_version is not None and bootloader is None:
        raise click.UsageError("Bootloader version with no image.")

    if debug_mode:
        display_debug_warning()
        # Default to no version checking
        if application_version_internal is None:
            application_version_internal=Package.DEFAULT_APP_VERSION
        if bootloader_version is None:
            bootloader_version=Package.DEFAULT_BL_VERSION
        if hw_version is None:
            hw_version=Package.DEFAULT_HW_VERSION
        if sd_req is None:
            # Use string as this will be mapped into an int below
            sd_req=str(Package.DEFAULT_SD_REQ[0])

    # Version checks
    if hw_version is None:
        raise click.UsageError("--hw-version required.")

    if sd_req is None and external_app is False:
        raise click.UsageError("--sd-req required.")

    if application is not None and application_version_internal is None:
        raise click.UsageError('--application-version or --application-version-string'
                   ' required with application image.')

    if bootloader is not None and bootloader_version is None:
        raise click.UsageError("--bootloader-version required with bootloader image.")

    # Zigbee only allows App, SoftDevice (minor), bootloader or Softdevice+bootloader
    if zigbee:
        if sum(bool(x) for x in [application, softdevice, bootloader]) != 1:
            click.echo('Error: Provide either --application, --softdevice, or --bootloader'
                       ' for Zigbee package generation (not a combination).')

    if application is not None and softdevice is not None and sd_id is None:
        raise click.UsageError("--sd-id required with softdevice and application images.")

    if application is None and external_app is True:
        raise click.UsageError("--external-app requires an application.")

    if application is not None and softdevice is not None and external_app is True:
        raise click.UsageError("--external-app is only possible for application only DFU packages.")

    if application is not None and bootloader is not None and external_app is True:
        raise click.UsageError("--external-app is only possible for application only DFU packages.")

    if zigbee and zigbee_ota_hw_version is None:
        raise click.UsageError("--zigbee-ota-hw-version is required.")

    if zigbee and zigbee_ota_fw_version is None:
        zigbee_ota_fw_version = 0

    sd_req_list = []
    if sd_req is not None:
        try:
            # This will parse any string starting with 0x as base 16.
            sd_req_list = sd_req.split(',')
            sd_req_list = list(map(int_as_text_to_int, sd_req_list))
        except ValueError:
            raise NordicSemiException("Could not parse value for --sd-req. "
                                      "Hex values should be prefixed with 0x.")

    sd_id_list = []
    if sd_id is not None:
        try:
            # This will parse any string starting with 0x as base 16.
            sd_id_list = sd_id.split(',')
            sd_id_list = list(map(int_as_text_to_int, sd_id_list))

            # Copy all IDs from sd_id_list to sd_req_list, without duplicates.
            # This ensures that the softdevice update can be repeated in case
            # SD+(BL)+App update terminates during application update after the
            # softdevice was already updated (with new ID). Such update would
            # have to be repeated and the softdevice would have to be sent again,
            # this time updating itself.
            sd_req_list += set(sd_id_list) - set(sd_req_list)
        except ValueError:
            raise NordicSemiException("Could not parse value for --sd-id. "
                                      "Hex values should be prefixed with 0x.")
    else:
        sd_id_list = sd_req_list

    if key_file is None:
        display_nokey_warning()
        signer = None
    else:
        signer = Signing()
        default_key = signer.load_key(key_file)
        if default_key:
            display_sec_warning()

    if zigbee_comment is None:
        zigbee_comment = ''
    elif any(ord(char) > 127 for char in zigbee_comment): # Check if all the characters belong to the ASCII range
        click.echo('Warning: Non-ASCII characters in the comment are not allowed. Discarding comment.')
        zigbee_comment = ''
    elif len(zigbee_comment) > 30:
        click.echo('Warning: truncating the comment to 30 bytes.')
        zigbee_comment = zigbee_comment[:30]

    if zigbee_manufacturer_id is None:
        zigbee_manufacturer_id = 0xFFFF

    if zigbee_image_type is None:
        zigbee_image_type = 0xFFFF

    # Set the external_app to false in --zigbee is set
    inner_external_app = external_app
    if zigbee:
        inner_external_app = False

    if zigbee_ota_min_hw_version is not None and zigbee_ota_min_hw_version > 0xFFFF:
        raise click.BadParameter('Exceeds 2-byte long integer.', param_hint='zigbee-ota-min-hw-version')

    if zigbee_ota_max_hw_version is not None and zigbee_ota_max_hw_version > 0xFFFF:
        raise click.BadParameter('Exceeds 2-byte long integer.', param_hint='zigbee-ota-max-hw-version')

    if zigbee and (hw_version > 0xFFFF):
        raise click.BadParameter('Exceeds 2-byte long integer.', param_hint='hw-version')

    # Warn user if minimal/maximum zigbee ota hardware version are not correct:
    #   * only one of them is given
    #   * minimum version is higher than maximum version
    #   * hw_version is inside the range specified by minimum and maximum hardware version
    if (type(zigbee_ota_min_hw_version) is int) != (type(zigbee_ota_max_hw_version) is int):
        click.echo('Warning: min/max zigbee ota hardware version is missing. Discarding min/max hardware version.')
    elif type(zigbee_ota_min_hw_version) is int:
        if zigbee_ota_min_hw_version > zigbee_ota_max_hw_version:
            click.echo('Warning: zigbee-ota-min-hw-version is higher than zigbee-ota-max-hw-version.')
        if (hw_version > zigbee_ota_max_hw_version) or (hw_version < zigbee_ota_min_hw_version):
            click.echo('Warning: hw-version is outside the specified range specified by zigbee_ota_min_hw_version and zigbee_ota_max_hw_version.')

    # Generate a DFU package. If --zigbee is set this is the inner DFU package
    # which will be used as a binary input to the outer DFU package
    package = Package(debug_mode,
                      hw_version,
                      application_version_internal,
                      bootloader_version,
                      sd_req_list,
                      sd_id_list,
                      application,
                      bootloader,
                      softdevice,
                      sd_boot_validation,
                      app_boot_validation,
                      signer,
                      inner_external_app,
                      zigbee,
                      zigbee_manufacturer_id,
                      zigbee_image_type,
                      zigbee_comment,
                      zigbee_ota_min_hw_version,
                      zigbee_ota_max_hw_version)

    package.generate_package(zipfile_path)

    if zigbee:
        from shutil import copyfile
        from os import remove

        log_message = "Zigbee update created at {0}".format(package.zigbee_ota_file.filename)
        click.echo(log_message)

        # Taking the inner Zigbee package as input for the outer DFU package
        binfile = package.zigbee_ota_file.filename.replace(".zigbee", ".bin")
        copyfile(package.zigbee_ota_file.filename, binfile)

        # Create the outer Zigbee DFU package.
        package = Package(debug_mode,
                          zigbee_ota_hw_version,
                          zigbee_ota_fw_version,
                          None,
                          [0xFFFE],
                          [0xFFFE],
                          binfile,
                          None,
                          None,
                          None,
                          None,
                          signer,
                          True)

        package.generate_package(zipfile_path)
        remove(binfile)

    log_message = "Zip created at {0}".format(zipfile_path)
    click.echo(log_message)

@pkg.command(short_help='Display the contents of a .zip package file.')
@click.argument('zip_file', required=True, type=click.Path())

def display(zip_file):

    package = Package()
    package.parse_package(zip_file, preserve_work_dir=True)

    click.echo("{0}".format(str(package)))

global_bar = None
def update_progress(progress=0):
    if global_bar:
        global_bar.update(progress)

@cli.group(short_help='Perform a Device Firmware Update over serial, BLE, Thread, Zigbee or ANT transport given a DFU package (zip file).')
def dfu():
    """
    This set of commands supports Device Firmware Upgrade procedures over both BLE and serial transports.
    """
    pass

def do_serial(package, port, connect_delay, flow_control, packet_receipt_notification, baud_rate, serial_number, ping,
              timeout):

    if flow_control is None:
        flow_control = DfuTransportSerial.DEFAULT_FLOW_CONTROL
    if packet_receipt_notification is None:
        packet_receipt_notification = DfuTransportSerial.DEFAULT_PRN
    if baud_rate is None:
        baud_rate = DfuTransportSerial.DEFAULT_BAUD_RATE
    if ping is None:
        ping = False
    if port is None:
        device_lister = DeviceLister()
        device = device_lister.get_device(serial_number=serial_number)
        if device is None:
            raise NordicSemiException("A device with serial number %s is not connected." % serial_number)
        port = device.get_first_available_com_port()
        logger.info("Resolved serial number {} to port {}".format(serial_number, port))

    if timeout is None:
        timeout = DfuTransportSerial.DEFAULT_TIMEOUT

    logger.info("Using board at serial port: {}".format(port))
    serial_backend = DfuTransportSerial(com_port=str(port), baud_rate=baud_rate,
                                        flow_control=flow_control, prn=packet_receipt_notification, do_ping=ping,
                                        timeout=timeout)
    serial_backend.register_events_callback(DfuEvent.PROGRESS_EVENT, update_progress)
    dfu = Dfu(zip_file_path = package, dfu_transport = serial_backend, connect_delay = connect_delay)

    if logger.getEffectiveLevel() > logging.INFO:
        with click.progressbar(length=dfu.dfu_get_total_size()) as bar:
            global global_bar
            global_bar = bar
            dfu.dfu_send_images()
    else:
        dfu.dfu_send_images()

    click.echo("Device programmed.")

@dfu.command(short_help='Update the firmware on a device over a USB serial connection. The DFU '
                        'target must be a chip with USB pins (i.e. nRF52840) and provide a USB ACM '
                        'CDC serial interface.')
@click.option('-pkg', '--package',
              help='Filename of the DFU package.',
              type=click.Path(exists=True, resolve_path=True, file_okay=True, dir_okay=False),
              required=True)
@click.option('-p', '--port',
              help='Serial port address to which the device is connected. (e.g. COM1 in windows systems, /dev/ttyACM0 in linux/mac)',
              type=click.STRING,
              cls = OptionRequiredIf)
@click.option('-cd', '--connect-delay',
              help='Delay in seconds before each connection to the target device during DFU. Default is 3.',
              type=click.INT,
              required=False)
@click.option('-fc', '--flow-control',
              help='To enable flow control set this flag to 1',
              type=click.BOOL,
              required=False)
@click.option('-prn', '--packet-receipt-notification',
              help='Set the packet receipt notification value',
              type=click.INT,
              required=False)
@click.option('-b', '--baud-rate',
              help='Set the baud rate',
              type=click.INT,
              required=False)
@click.option('-snr', '--serial-number',
              help='Serial number of the device. Ignored if --port is set.',
              type=click.STRING,
              required=False)
@click.option('-t', '--timeout',
              help='Set the timeout in seconds for board to respond (default: 30 seconds)',
              type=click.INT,
              required=False)
def usb_serial(package, port, connect_delay, flow_control, packet_receipt_notification, baud_rate, serial_number,
               timeout):
    """Perform a Device Firmware Update on a device with a bootloader that supports USB serial DFU."""
    do_serial(package, port, connect_delay, flow_control, packet_receipt_notification, baud_rate, serial_number, False,
              timeout)


@dfu.command(short_help="Update the firmware on a device over a UART serial connection. The DFU target must be a chip using digital I/O pins as an UART.")
@click.option('-pkg', '--package',
              help='Filename of the DFU package.',
              type=click.Path(exists=True, resolve_path=True, file_okay=True, dir_okay=False),
              required=True)
@click.option('-p', '--port',
              help='Serial port address to which the device is connected. (e.g. COM1 in windows systems, /dev/ttyACM0 in linux/mac)',
              type=click.STRING,
              cls = OptionRequiredIf)
@click.option('-cd', '--connect-delay',
              help='Delay in seconds before each connection to the target device during DFU. Default is 3.',
              type=click.INT,
              required=False)
@click.option('-fc', '--flow-control',
              help='To enable flow control set this flag to 1',
              type=click.BOOL,
              required=False)
@click.option('-prn', '--packet-receipt-notification',
              help='Set the packet receipt notification value',
              type=click.INT,
              required=False)
@click.option('-b', '--baud-rate',
              help='Set the baud rate',
              type=click.INT,
              required=False)
@click.option('-snr', '--serial-number',
              help='Serial number of the device. Ignored if --port is set.',
              type=click.STRING,
              required=False)
@click.option('-t', '--timeout',
              help='Set the timeout in seconds for board to respond (default: 30 seconds)',
              type=click.INT,
              required=False)
def serial(package, port, connect_delay, flow_control, packet_receipt_notification, baud_rate, serial_number,
           timeout):
    """Perform a Device Firmware Update on a device with a bootloader that supports UART serial DFU."""

    do_serial(package, port, connect_delay, flow_control, packet_receipt_notification, baud_rate, serial_number, True,
              timeout)


def enumerate_ports():
    device_lister = DeviceLister()
    descs = device_lister.enumerate()
    if len(descs) == 0:
        raise click.UsageError("\nNo boards found.")

    click.echo('Please select connectivity serial port:')
    for i, choice in enumerate(descs):
        click.echo('\t{} : {} - {}'.format(
                                        i,
                                        choice.get_first_available_com_port(),
                                        choice.serial_number))

    index = click.prompt('Enter your choice: ',
                         type=click.IntRange(0, len(descs)))
    device = descs[index]
    is_jlink = device.vendor_id == "1366"
    return device.get_first_available_com_port(), is_jlink


def get_port_by_snr(snr):
    device_lister = DeviceLister()
    device = device_lister.get_device(serial_number=snr)
    if not device:
        raise NordicSemiException('Board not found')
    is_jlink = device.vendor_id == "1366"
    return device.get_first_available_com_port(), is_jlink


def port_is_jlink(port):
    device_lister = DeviceLister()
    device = device_lister.get_device(com=port)
    if not device:
        raise NordicSemiException('Board not found')
    return device.vendor_id == "1366"


@dfu.command(short_help="Update the firmware on a device over a BLE connection.")
@click.option('-pkg', '--package',
              help='Filename of the DFU package.',
              type=click.Path(exists=True, resolve_path=True, file_okay=True, dir_okay=False),
              required=True)
@click.option('-ic', '--conn-ic-id',
              help='Connectivity IC family: NRF51 or NRF52',
              type=click.Choice(['NRF51', 'NRF52']),
              required=True)
@click.option('-p', '--port',
              help='Serial port COM port to which the connectivity IC is connected.',
              type=click.STRING)
@click.option('-cd', '--connect-delay',
              help='Delay in seconds before each connection to the target device during DFU. Default is 3.',
              type=click.INT,
              required=False)
@click.option('-n', '--name',
              help='Device name.',
              type=click.STRING)
@click.option('-a', '--address',
              help='BLE address of the DFU target device.',
              type=click.STRING)
@click.option('-snr', '--jlink_snr',
              help='Jlink serial number for the connectivity IC.',
              type=click.STRING)
@click.option('-f', '--flash_connectivity',
              help='Flash connectivity firmware automatically. Default: disabled.',
              type=click.BOOL,
              is_flag=True)
@click.option('-mtu', '--att-mtu',
              help='ATT MTU. Maximum ATT packet size for BLE transfers. '
                   'Accepted values in range [23, 247]. Default is 247. '
                   'Note: Failing DFU transmissions can in some cases be solved by setting a '
                   'lower mtu.',
              type=click.IntRange(23, 247, clamp=True),
              default=247)
def ble(package, conn_ic_id, port, connect_delay, name, address, jlink_snr, flash_connectivity, att_mtu):
    """
    Perform a Device Firmware Update on a device with a bootloader that supports BLE DFU.
    This requires a second nRF device, connected to this computer, with connectivity firmware
    loaded. The connectivity device will perform the DFU procedure onto the target device.
    """
    ble_driver_init(conn_ic_id)
    if name is None and address is None:
        name = 'DfuTarg'
        click.echo("No target selected. Default device name: {} is used.".format(name))

    # Remove colons from address in case written in format XX:XX:XX:XX:XX:XX
    if address:
        address = address.replace(':', '')
        if not re.match('^[0-9A-Fa-f]{12}$', address):
            raise click.BadParameter('Must be exactly 6 bytes HEX, '
                                     'e.g. ABCDEF123456 or AB:CD:EF:12:34:56.', param_hint='address')

    if port is None and jlink_snr is not None:
        port, is_jlink = get_port_by_snr(jlink_snr)
    elif port is None:
        port, is_jlink = enumerate_ports()
    else:
        is_jlink = port_is_jlink(port)

    if flash_connectivity:
        if is_jlink:
            flasher = Flasher(serial_port=port, snr=jlink_snr)
            if flasher.fw_check():
                click.echo("Board already flashed with connectivity firmware.")
            else:
                click.echo("Flashing connectivity firmware...")
                flasher.fw_flash()
                click.echo("Connectivity firmware flashed.")
            flasher.reset()
            time.sleep(1)
        else:
            click.echo("Flashing connectivity firmware...")
            serial_backend = DfuTransportSerial(com_port=str(port))
            serial_backend.register_events_callback(DfuEvent.PROGRESS_EVENT,
                                                    update_progress)
            connectivity_firmware = os.path.join(
                                        os.path.dirname(config.__file__),
                                        "hex",
                                        "sd_api_v5",
                                        "connectivity_4.1.4_usb_with_s132_5.1.0_dfu_pkg.zip"
                                    )
            dfu = Dfu(zip_file_path=connectivity_firmware,
                      dfu_transport=serial_backend,
                      connect_delay=connect_delay)
            dfu.dfu_send_images()
            click.echo("Connectivity firmware flashed.")

    logger.info("Using connectivity board at serial port: {}".format(port))
    ble_backend = DfuTransportBle(serial_port=str(port),
                                  att_mtu=att_mtu,
                                  target_device_name=str(name),
                                  target_device_addr=str(address))
    ble_backend.register_events_callback(DfuEvent.PROGRESS_EVENT, update_progress)
    dfu = Dfu(zip_file_path=package, dfu_transport=ble_backend, connect_delay=connect_delay)

    if logger.getEffectiveLevel() > logging.INFO:
        with click.progressbar(length=dfu.dfu_get_total_size()) as bar:
            global global_bar
            global_bar = bar
            dfu.dfu_send_images()
    else:
        dfu.dfu_send_images()

    click.echo("Device programmed.")


@dfu.command(short_help="Update the firmware on a device over an ANT connection.")
@click.option('-pkg', '--package',
              help='Filename of the DFU package.',
              type=click.Path(exists=True, resolve_path=True, file_okay=True, dir_okay=False),
              required=True)
@click.option('-p', '--port',
              help='ANT USB device to use for performing the update.',
              type=click.INT,
              required=False)
@click.option('-cd', '--connect-delay',
              help='Delay in seconds before each connection to the target device during DFU. Default is 3.',
              type=click.INT,
              required=False)
@click.option('-prn', '--packet-receipt-notification',
              help='Set the packet receipt notification value.',
              type=click.INT,
              required=False)
@click.option('--period',
              help='Set the ANT Channel period.',
              type=click.INT,
              required=False)
@click.option('--freq',
              help='Set the ANT RF Frequency.',
              type=click.INT,
              required=False)
@click.option('--net-key',
              help='Set the ANT network key. Must be formatted as hexadecimal numbers separated by dashes ("-").',
              type=click.STRING,
              required=False)
@click.option('--dev-type',
              help='Set the ANT device type',
              type=click.INT,
              required=False)
@click.option('-srn', '--serial',
              help='Serial number of device to search for.',
              type=click.INT,
              required=False)
@click.option('-d', '--debug/--no-debug',
              help='Enable ANT debug logs.',
              default=False,
              required=False)
def ant(package, port, connect_delay, packet_receipt_notification, period,
        freq, net_key, dev_type, serial, debug):

    from nordicsemi.dfu.dfu_transport_ant import platform_supported

    if not platform_supported():
        return

    # This import needs to happen only if the platform is supported.
    from nordicsemi.dfu.dfu_transport_ant import DfuTransportAnt, AntParams

    ant_config = AntParams()
    if port is None:
        port = DfuTransportAnt.DEFAULT_PORT
    if packet_receipt_notification is None:
        packet_receipt_notification = DfuTransportAnt.DEFAULT_PRN
    if period is not None:
        ant_config.channel_period = period
    if freq is not None:
        ant_config.rf_freq = freq
    if net_key is not None:
        ant_config.network_key = [int(x, 16) for x in net_key.split('-')]
    if dev_type is not None:
        ant_config.device_type = dev_type
    if serial is not None:
        ant_config.device_num = serial & 0xFFFF
        ant_config.trans_type = 0x01 | ((serial >> 12) & 0xF0)

    ant_backend = DfuTransportAnt(port=port, prn=packet_receipt_notification,
        ant_config=ant_config, debug=debug)
    ant_backend.register_events_callback(DfuEvent.PROGRESS_EVENT, update_progress)
    dfu = Dfu(zip_file_path=package, dfu_transport=ant_backend, connect_delay=connect_delay)

    try:
        if logger.getEffectiveLevel() > logging.INFO:
            with click.progressbar(length=dfu.dfu_get_total_size()) as bar:
                global global_bar
                global_bar = bar
                dfu.dfu_send_images()
        else:
            dfu.dfu_send_images()
    except Exception:
        if ant_backend.dfu_adapter and ant_backend.dfu_adapter.ant_dev:
            # Make sure things get cleaned up if there is an error.
            ant_backend.dfu_adapter.ant_dev.ant_close()
        raise

    click.echo("Device programmed.")


def convert_version_string_to_int(s):
    """Convert from semver string "1.2.3", to integer 10203"""
    numbers = s.split(".")
    if len(numbers) != 3:
        raise click.BadParameter("Must be on the format x.y.z", param_hint='application-version-string')
    js = [10000, 100, 1]
    return sum([js[i] * int(numbers[i]) for i in range(3)])

@dfu.command(short_help="Update the firmware on a device over a Thread connection.")
@click.option('-pkg', '--package',
              help='Filename of the DFU package.',
              type=click.Path(exists=True, resolve_path=True, file_okay=True, dir_okay=False),
              required=True)
@click.option('-p', '--port',
              help='Serial port COM port to which the NCP is connected.',
              type=click.STRING)
@click.option('-a', '--address',
              help='Device IPv6 address. If address is not specified then perform DFU '
                   'on all capable devices. If multicast address is specified (FF03::1), '
                   'perform multicast DFU.',
              type=click.STRING)
@click.option('-sp', '--server_port',
              help='UDP port to which the DFU server binds. If not specified the 5683 is used.',
              type=click.INT,
              default=5683)
@click.option('--panid',
              help='802.15.4 PAN ID. If not specified then 1234 is used as PAN ID.',
              type=click.INT)
@click.option('--channel',
              help='802.15.4 Channel. If not specified then channel 11 is used.',
              type=click.INT)
@click.option('-snr', '--jlink_snr',
              help='Jlink serial number.',
              type=click.STRING)
@click.option('-f', '--flash_connectivity',
              help='Flash NCP connectivity firmware automatically. Default: disabled.',
              type=click.BOOL,
              is_flag=True)
@click.option('-s', '--sim',
              help='Use software NCP and connect to the OT simulator.',
              type=click.BOOL,
              is_flag=True)
@click.option('-r', '--rate',
              help="Multicast upload rate in blocks per second.",
              type=click.FLOAT)
@click.option('-rs', '--reset_suppress',
              help='Suppress device reset after finishing DFU for a given number of milliseconds. ' +
                   'If -1 is given then suppress indefinatelly.',
              type = click.INT,
              metavar = '<delay_in_ms>')
@click.option('-m', '--masterkey',
              help='Masterkey. If not specified then 00112233445566778899aabbccddeeff is used',
              type=click.STRING)

def thread(package, port, address, server_port, panid, channel, jlink_snr, flash_connectivity,
           sim, rate, reset_suppress, masterkey):
    """
    Perform a Device Firmware Update on a device that supports Thread DFU.
    This requires a second nRF device, connected to this computer, with Thread Network
    CoProcessor (NCP) firmware loaded. The NCP device will perform the DFU procedure onto
    the target device.
    """
    ble_driver_init('NRF52')
    from nordicsemi.thread import tncp
    from nordicsemi.thread.dfu_thread import create_dfu_server
    from nordicsemi.thread.tncp import NCPTransport
    from nordicsemi.thread.ncp_flasher import NCPFlasher

    mcast_dfu = False

    if address is None:
        address = ipaddress.ip_address("ff03::1")
        click.echo("Address not specified. Using ff03::1 (all Thread nodes)")
    else:
        try:
            address = ipaddress.ip_address(address)
            mcast_dfu = address.is_multicast
        except:
            click.echo("Invalid IPv6 address")
            return 1

    if (not sim):
        if port is None and jlink_snr is None:
            click.echo("Please specify serial port or Jlink serial number.")
            return 2

        elif port is None:
            port, is_jlink = get_port_by_snr(jlink_snr)
            if port is None or not is_jlink:
                click.echo("\nNo Segger USB CDC ports found, please connect your board.")
                return 3

        stream_descriptor = 'u:' + port
        click.echo("Using connectivity board at serial port: {}".format(port))
    else:
        stream_descriptor = 'p:' + Flasher.which('ot-ncp') + ' 30'
        click.echo("Using ot-ncp binary: {}".format(stream_descriptor))

    if flash_connectivity:
        flasher = NCPFlasher(serial_port=port, snr = jlink_snr)
        if flasher.fw_check():
            click.echo("Board already flashed with connectivity firmware.")
        else:
            click.echo("Flashing connectivity firmware...")
            flasher.fw_flash()
            click.echo("Connectivity firmware flashed.")

        flasher.reset()

        # Delay is required because NCP needs time to initialize.
        time.sleep(1.0)

    config = tncp.NCPTransport.get_default_config()
    if (panid):
        config[tncp.NCPTransport.CFG_KEY_PANID] = panid
    if (channel):
        config[tncp.NCPTransport.CFG_KEY_CHANNEL] = channel
    if (flash_connectivity):
        config[tncp.NCPTransport.CFG_KEY_RESET] = False
    if (masterkey):
        config[tncp.NCPTransport.CFG_KEY_MASTERKEY] = util.hex_to_bytes(masterkey)

    opts = type('DFUServerOptions', (object,), {})()
    opts.rate = rate
    opts.reset_suppress = reset_suppress
    opts.mcast_dfu = mcast_dfu

    transport = NCPTransport(server_port, stream_descriptor, config)
    dfu = create_dfu_server(transport, package, opts)

    try:
        sighandler = lambda signum, frame : transport.close()
        signal.signal(signal.SIGINT, sighandler)
        signal.signal(signal.SIGTERM, sighandler)

        transport.open()
        # Delay DFU trigger until NCP promotes to a router (6 seconds by default)
        click.echo("Waiting for NCP to promote to a router...")
        time.sleep(6.0)
        dfu.trigger(address, 3)
        click.echo("Thread DFU server is running... Press <Ctrl + D> to stop.")
        pause()
        click.echo("Terminating")

    except Exception as e:
        logger.exception(e)
    finally:
        transport.close()

@dfu.command(short_help="Update the firmware on a device over a Zigbee connection.")
@click.option('-f', '--file',
              help='Filename of the Zigbee OTA Upgrade file.',
              type=click.Path(exists=True, resolve_path=True, file_okay=True, dir_okay=False),
              required=True)
@click.option('-snr', '--jlink_snr',
              help='JLink serial number of the devboard which shall serve as a OTA Server cluster',
              type=click.STRING)
@click.option('-chan', '--channel',
              help='802.15.4 Channel that the OTA server will use',
              type=click.INT)

def zigbee(file, jlink_snr, channel):
    """
    Perform a Device Firmware Update on a device that implements a  Zigbee OTA Client cluster.
    This requires a second nRF device, connected to this computer, which shall serve as a
    OTA Server cluster.
    """
    ble_driver_init('NRF52')
    from nordicsemi.zigbee.ota_flasher import OTAFlasher
    of = OTAFlasher(fw = file, channel = channel, snr = jlink_snr)

    if of.fw_check():
        click.echo("Board already flashed with connectivity firmware.")
    else:
        click.echo("Flashing connectivity firmware...")
        of.fw_flash()
        click.echo("Connectivity firmware flashed.")

    of.reset()
    time.sleep(3.0) # A delay to init the OTA Server flashed on the devboard and the CLI inside of it
    of.randomize_eui64()
    of.setup_channel()

@cli.group()
def zigbee():
    """
    Zigbee-related commands and utilities.
    """
    pass


def _pretty_help_option(text: str):
    formatted_lines = []
    for line in text.split("\n"):
        formatted_lines.append(line + " " * 100)
    return "\n".join(formatted_lines)


@zigbee.command(short_help='Generate the Zigbee Production Config (version 1) hex file.',
                name='production_config',)
@click.argument('input', required=True, type=click.Path())
@click.argument('output', required=True, type=click.Path())
@click.option('--offset', type=BASED_INT_OR_NONE, help=_pretty_help_option(
    "Offset at which the Production Config is located.\n"
    "Depending on the SDK and the device versions, use the following values:\n"
    f"{ProductionConfig.offset_help()}"
    f"By default, the value for {ProductionConfig.DEFAULT_OFFSET_SDK} "
    f"{ProductionConfig.DEFAULT_OFFSET_CHIP} is used."))
def production_config(input, output, offset):
    """
    Generate the Production config hex file for Zigbee Devices out of YAML-structured description.
    Generated Production config is in version 1.

    INPUT - path to yaml file.\n
            Example yaml content:

    \b
                channel_mask: 0x00100000
                install_code: 83FED3407A939723A5C639B26916D505
                extended_address: AABBCCDDEEFF0011
                tx_power: 9
                app_data: 01ABCD

    OUTPUT - name of output file
    """
    try:
        pc = ProductionConfig(input)
    except ProductionConfigWrongException:
        raise click.UsageError("Input YAML file format wrong."
                               " Please see the example YAML file in the documentation.")

    try:
        if offset is None:
            pc.generate(output)
        else:
            pc.generate(output, offset=offset)
        click.echo("Production Config hexfile generated.")
    except ProductionConfigTooLargeException as e:
        raise click.UsageError(f"Production Config too large: {e.length} bytes")


if __name__ == '__main__':
    cli()



#
# Copyright (c) 2018 Nordic Semiconductor ASA
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#   1. Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright notice, this
#   list of conditions and the following disclaimer in the documentation and/or
#   other materials provided with the distribution.
#
#   3. Neither the name of Nordic Semiconductor ASA nor the names of other
#   contributors to this software may be used to endorse or promote products
#   derived from this software without specific prior written permission.
#
#   4. This software must only be used in or with a processor manufactured by Nordic
#   Semiconductor ASA, or in or with a processor manufactured by a third party that
#   is used in combination with a processor manufactured by Nordic Semiconductor.
#
#   5. Any software provided in binary or object form under this license must not be
#   reverse engineered, decompiled, modified and/or disassembled.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import crcmod.predefined
import struct
import intelhex
import yaml


class ProductionConfigWrongException(Exception):
    pass


class ProductionConfigTooLargeException(Exception):
    def __init__(self, length):
        self.length = length


class ProductionConfig:
    SIZE_MAX = 128
    MAGIC_NUMBER = 0xF6DD37E7
    VERSION = 1
    HEADER_FORMAT = '<HHLQ16s16sH'
    OFFSETS = {
        "SDK 3.x": {
            "nRF52840": 0xFC000,
        },
        "SDK 4.x": {
            "nRF52840": 0xFF000,
            "nRF52833": 0x7F000,
        }
    }
    DEFAULT_OFFSET_SDK = "SDK 3.x"
    DEFAULT_OFFSET_CHIP = "nRF52840"
    DEFAULT_OFFSET = OFFSETS[DEFAULT_OFFSET_SDK][DEFAULT_OFFSET_CHIP]
    DEFAULT_CHANNEL = 0x07FFF800

    @classmethod
    def offset_help(cls):
        return format_offsets(cls.OFFSETS)

    def __init__(self, path):
        self._parsed_values = {}
        self._crc16 = crcmod.predefined.mkPredefinedCrcFun('x-25')

        try:
            # Open the YAML file
            with open(path, 'r') as f:
                self._yaml = yaml.load(f, Loader=yaml.FullLoader)

        except yaml.YAMLError as e:
            raise ProductionConfigWrongException

        try:
            # Handle the channel mask
            if "channel_mask" not in self._yaml:
                self._parsed_values["channel_mask"] = self.DEFAULT_CHANNEL
            else:
                self._parsed_values["channel_mask"] = self._yaml["channel_mask"]

            # Handle the EUI64 extended address
            if "extended_address" not in self._yaml:
                self._parsed_values["extended_address"] = 0  # int('0000000000000000', 16)
            else:
                self._parsed_values["extended_address"] = int(self._yaml["extended_address"], 16)

            # Handle the Install code
            if "install_code" not in self._yaml:
                self._parsed_values["install_code"] = bytes(16)
                self._ic_crc = 0
            else:
                self._parsed_values["install_code"] = bytes.fromhex(self._yaml["install_code"])
                self._ic_crc = self._crc16(self._parsed_values["install_code"])

            # Handle the Transmission Power
            if "tx_power" not in self._yaml:
                self._parsed_values["tx_power"] = bytes(16)
            else:
                self._parsed_values["tx_power"] = bytes([self._yaml["tx_power"]] * 16)

            # Handle Application Data (optional)
            if "app_data" not in self._yaml:
                self._parsed_values["app_data"] = b''
                self._ad_len = 0
            else:
                self._parsed_values["app_data"] = bytes.fromhex(self._yaml["app_data"])
                self._ad_len = len(self._parsed_values["app_data"])

        except (TypeError, ValueError) as e:
            raise ProductionConfigWrongException

    def _custom_crc32(self, data):
        '''
            Adapted from C routine inside the Zigbee stack - custom CRC32 that guards the Production Config
        '''
        ZB_CRC32_POLY = 0x04C11DB7
        crc = 0
        for d in data:
            c  = ((( crc ^ d ) & 0xff) << 24)
            for j in range(8):
                if c & 0x80000000:
                    c = (c << 1) ^ ZB_CRC32_POLY
                else:
                    c = (c << 1)
            crc = (0xFFFFFFFF & ((crc >> 8) ^ c))
        return (~crc & 0xFFFFFFFF)

    def generate(self, path, offset=DEFAULT_OFFSET):
        # Calculate the CRC-16 of the install code
        self._struct = (struct.pack(self.HEADER_FORMAT,
                                    struct.calcsize(self.HEADER_FORMAT) + 4 + self._ad_len,  # Plus the CRC-32; plus the app_data
                                    self.VERSION,
                                    self._parsed_values["channel_mask"],
                                    self._parsed_values["extended_address"],
                                    self._parsed_values["tx_power"],
                                    self._parsed_values["install_code"],
                                    self._ic_crc) +
                                   self._parsed_values["app_data"])

        crc32 = self._custom_crc32(self._struct)

        output = struct.pack('<L', self.MAGIC_NUMBER) + struct.pack('<L', crc32) + self._struct

        if len(output) > self.SIZE_MAX + 4: # 4 is for Magic Number
            raise ProductionConfigTooLargeException(len(output))

        ih = intelhex.IntelHex()
        ih.puts(offset, output)
        ih.write_hex_file(path)


def format_offsets(offset_dict: dict):
    result = ""
    for sdk, boards in offset_dict.items():
        result += f"{sdk}:\n"
        for board, offset in boards.items():
            result += f"- {board}: {hex(offset)}\n"
    return result



#
# Copyright (c) 2016 Nordic Semiconductor ASA
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#   1. Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright notice, this
#   list of conditions and the following disclaimer in the documentation and/or
#   other materials provided with the distribution.
#
#   3. Neither the name of Nordic Semiconductor ASA nor the names of other
#   contributors to this software may be used to endorse or promote products
#   derived from this software without specific prior written permission.
#
#   4. This software must only be used in or with a processor manufactured by Nordic
#   Semiconductor ASA, or in or with a processor manufactured by a third party that
#   is used in combination with a processor manufactured by Nordic Semiconductor.
#
#   5. Any software provided in binary or object form under this license must not be
#   reverse engineered, decompiled, modified and/or disassembled.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
import tempfile
import os.path
import logging
import piccata

from nordicsemi.dfu.package import Package
from nordicsemi.thread.dfu_server import ThreadDfuServer

logger = logging.getLogger(__name__)

def _get_manifest_items(manifest):
    import inspect
    result = []

    for key, value in inspect.getmembers(manifest):
        if (key.startswith('__')):
            continue
        if not value:
            continue
        if inspect.ismethod(value) or inspect.isfunction(value):
            continue

        result.append((key, value))

    return result

def _get_file_names(manifest):
    data_attrs = _get_manifest_items(manifest)
    if (len(data_attrs) > 1):
        raise RuntimeError("More than one image present in manifest")
    data_attrs = data_attrs[0]
    firmware = data_attrs[1]
    logger.info("Image type {} found".format(data_attrs[0]))
    return firmware.dat_file, firmware.bin_file

def create_dfu_server(transport, zip_file_path, opts):
    '''
    Create a DFU server instance.
    :param transpoort: A transport to be used.
    :param zip_file_path: A path to the firmware package.
    :param opts: Optional parameters:
        mcast_dfu: An information if multicast DFU is enabled.
        rate: Multicast block transfer rate, in blocks per second
        reset_suppress: A delay before sending multicast reset command (in milliseconds). -1 means that no reset will be sent.
    '''
    temp_dir = tempfile.mkdtemp(prefix="nrf_dfu_")
    unpacked_zip_path = os.path.join(temp_dir, 'unpacked_zip')
    manifest = Package.unpack_package(zip_file_path, unpacked_zip_path)

    protocol = piccata.core.Coap(transport)
    transport.register_receiver(protocol)

    init_file, image_file = _get_file_names(manifest)

    with open(os.path.join(unpacked_zip_path, init_file), 'rb') as f:
        init_data = f.read()
    with open(os.path.join(unpacked_zip_path, image_file), 'rb') as f:
        image_data = f.read()

    return ThreadDfuServer(protocol, init_data, image_data, opts)



#
# Copyright (c) 2019 Nordic Semiconductor ASA
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#   1. Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright notice, this
#   list of conditions and the following disclaimer in the documentation and/or
#   other materials provided with the distribution.
#
#   3. Neither the name of Nordic Semiconductor ASA nor the names of other
#   contributors to this software may be used to endorse or promote products
#   derived from this software without specific prior written permission.
#
#   4. This software must only be used in or with a processor manufactured by Nordic
#   Semiconductor ASA, or in or with a processor manufactured by a third party that
#   is used in combination with a processor manufactured by Nordic Semiconductor.
#
#   5. Any software provided in binary or object form under this license must not be
#   reverse engineered, decompiled, modified and/or disassembled.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#


import os
import sys
import ctypes
from importlib import import_module
import logging

from pc_ble_driver_py.exceptions import NordicSemiException


LIBUSB_ENDPOINT_IN = 0x80
LIBUSB_ENDPOINT_OUT = 0x00
LIBUSB_REQUEST_TYPE_STANDARD = 0x00 << 5
LIBUSB_REQUEST_TYPE_CLASS = 0x01 << 5
LIBUSB_REQUEST_TYPE_VENDOR = 0x02 << 5
LIBUSB_REQUEST_TYPE_RESERVED = 0x03 << 5
LIBUSB_RECIPIENT_DEVICE = 0x00
LIBUSB_RECIPIENT_INTERFACE = 0x01
LIBUSB_RECIPIENT_ENDPOINT = 0x02
LIBUSB_RECIPIENT_OTHER = 0x03

ReqTypeInterfaceClass = LIBUSB_REQUEST_TYPE_CLASS | LIBUSB_RECIPIENT_INTERFACE
ReqTypeIN = ReqTypeInterfaceClass | LIBUSB_ENDPOINT_IN
ReqTypeOUT = ReqTypeInterfaceClass | LIBUSB_ENDPOINT_OUT
NORDIC_SEM_VER_REQUEST = 8
NORDIC_DFU_INFO_REQUEST = 7
DFU_DETACH_REQUEST = 0

logger = logging.getLogger(__name__)

is_32_bit = ctypes.sizeof(ctypes.c_void_p) == 4
abs_file_dir = os.path.dirname(os.path.abspath(__file__))
rel_import_dir = ""

dfu_path = os.path.join("nordicsemi", "dfu")
if is_32_bit:
    libusb_path = os.path.join("libusb", "x86")
    abs_file_dir = abs_file_dir.replace(dfu_path, libusb_path)
    rel_import_dir = os.path.join(".", libusb_path)
else:
    libusb_path = os.path.join("libusb", "x64")
    abs_file_dir = abs_file_dir.replace(dfu_path, libusb_path)
    rel_import_dir = os.path.join(".", libusb_path)

for path in ['PATH', 'LD_LIBRARY_PATH', 'DYLD_LIBRARY_PATH', 'DYLD_FALLBACK_LIBRARY_PATH']:
    if path not in os.environ:
        os.environ[path] = ""
    os.environ[path] = rel_import_dir + os.pathsep + abs_file_dir + os.pathsep + os.environ[path]

class DFUTrigger:
    def __init__(self):
        self.context = None

        if sys.platform == 'win32':
            try:
                # Load the libusb dll in advance so that libusb1 module is better able to load it
                ctypes.CDLL(os.path.join(abs_file_dir, "libusb-1.0.dll"));
            except OSError as err:
                logger.info(err)

        try:
            self.usb1 = import_module('usb1')
            self.context = self.usb1.USBContext()
        except OSError as err:
            if "libusb" in str(err):
                if sys.platform == 'win32' or sys.platform == 'darwin':
                    show_msg = "Libusb1 binaries are bundled with nrfutil for Windows and MacOS. " \
                               "Python is unable to locate or load the binaries. "
                elif 'linux' in sys.platform:
                    show_msg = "Libusb1 binaries are bundled with some linux distributions. " \
                                    "If you see this message, they are probably not installed on your system. "\
                                    "If you want to use DFU trigger, please install 'libusb1' using your package manager. "\
                                    "E.g: 'sudo apt-get install libusb-dev'."
                else:
                    show_msg = "Libusb1 is not compatible with your operating system."

                logger.warning("Could not load libusb1-0 library, which is a requirement to use DFU trigger. "\
                            "This is not a problem unless you intend to use this functionality. "\
                            "{}".format(show_msg))
    def clean(self):
        self.usb1 = None
        if self.context:
            self.context.close()

    def select_device(self, listed_device):
        all_devices = self.context.getDeviceList()
        filtered_devices = [dev for dev in all_devices
            if hex(dev.getVendorID())[2:].lower() == listed_device.vendor_id.lower() and
            hex(dev.getProductID())[2:].lower() == listed_device.product_id.lower()]

        access_error = False
        triggerless_devices = 0

        for nordic_device in filtered_devices:
            try:
                handle = nordic_device.open()
                SNO = handle.getSerialNumber()
                handle.close()
                if (SNO.lower() == listed_device.serial_number.lower()):
                    return nordic_device

            except self.usb1.USBErrorNotFound as err:
                #  Devices with matching VID and PID as target, but without a trigger iface.
                triggerless_devices += 1
            except self.usb1.USBErrorAccess as err:
                access_error = True
            except self.usb1.USBErrorNotSupported as err:
                pass #  Unsupported device. Moving on

        if triggerless_devices > 0:
            logger.debug("DFU trigger: Could not find trigger interface for device with serial number {}. "\
            "{}/{} devices with same VID/PID were missing a trigger interface."\
            .format(listed_device.serial_number, triggerless_devices, len(filtered_devices)))

        if access_error:
            raise NordicSemiException("LIBUSB_ERROR_ACCESS: Unable to connect to trigger interface.")

    def get_dfu_interface_num(self, libusb_device):
        for setting in libusb_device.iterSettings():
            if setting.getClass() == 255 and \
            setting.getSubClass() == 1 and \
            setting.getProtocol() == 1:
                return setting.getNumber()

    def no_trigger_exception(self, device):
        return NordicSemiException("No trigger interface found for device with serial number: {}, Product ID: 0x{} and Vendor ID: 0x{}\n"
        .format(device.serial_number, device.product_id, device.vendor_id))

    def enter_bootloader_mode(self, listed_device):
        if self.context is None:
            raise NordicSemiException("No Libusb1 context found, but is required to use DFU trigger. " \
                                    "This likely happens because the libusb1-0 binaries are missing from your system, "\
                                    "or Python is unable to locate them.")
        libusb_device = self.select_device(listed_device)
        if libusb_device is None:
            raise self.no_trigger_exception(listed_device)
        device_handle = libusb_device.open()
        dfu_iface = self.get_dfu_interface_num(libusb_device)

        if dfu_iface is None:
            raise self.no_trigger_exception(listed_device)

        with device_handle.claimInterface(dfu_iface):
            arr = bytearray("0", 'utf-8')
            try:
                device_handle.controlWrite(ReqTypeOUT, DFU_DETACH_REQUEST, 0, dfu_iface, arr)
            except Exception as err:
                if "LIBUSB_ERROR_PIPE" in str(err):
                    return
        raise NordicSemiException("A diconnection event from libusb is expected when the usb device restarts after triggering bootloder. "\
        "The event was not received. This can be an indication that the device was unable to leave application mode. "\
        "Serial number: {}, Product ID: 0x{}, Vendor ID: 0x{}\n\n"
        .format(listed_device.serial_number, listed_device.product_id, listed_device.vendor_id))



#
# Copyright (c) 2016 Nordic Semiconductor ASA
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#   1. Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright notice, this
#   list of conditions and the following disclaimer in the documentation and/or
#   other materials provided with the distribution.
#
#   3. Neither the name of Nordic Semiconductor ASA nor the names of other
#   contributors to this software may be used to endorse or promote products
#   derived from this software without specific prior written permission.
#
#   4. This software must only be used in or with a processor manufactured by Nordic
#   Semiconductor ASA, or in or with a processor manufactured by a third party that
#   is used in combination with a processor manufactured by Nordic Semiconductor.
#
#   5. Any software provided in binary or object form under this license must not be
#   reverse engineered, decompiled, modified and/or disassembled.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import os
import struct

import unittest
from nordicsemi.dfu.bl_dfu_sett import BLDFUSettings
from nordicsemi.dfu.nrfhex import nRFArch
from nordicsemi.dfu.signing import Signing


class TestBLDFUSettingsV1(unittest.TestCase):
    def setUp(self):
        script_abspath = os.path.abspath(__file__)
        script_dirname = os.path.dirname(script_abspath)
        os.chdir(script_dirname)

    def test_fromhexfile(self):
        settings = BLDFUSettings()
        settings.fromhexfile('firmwares/bl_settings_v1_nrf52.hex')

        self.assertEqual(nRFArch.NRF52, settings.arch)
        self.assertEqual('nRF52', settings.arch_str)
        self.assertEqual(0x0007F000, settings.bl_sett_addr)
        self.assertEqual(0xF6744851, settings.crc)
        self.assertEqual(0x00000001, settings.bl_sett_ver)
        self.assertEqual(0x00000003, settings.app_ver)
        self.assertEqual(0x00000003, settings.bl_ver)
        self.assertEqual(0x00000000, settings.bank_layout)
        self.assertEqual(0x00000000, settings.bank_current)
        self.assertEqual(0x0002C160, settings.app_sz)
        self.assertEqual(0x62C83F81, settings.app_crc)
        self.assertEqual(0x00000001, settings.bank0_bank_code)

    def test_generate_without_application_file(self):
        settings = BLDFUSettings()
        settings.generate(arch='NRF52',
                          app_file=None,
                          app_ver=1,
                          bl_ver=1,
                          bl_sett_ver=1,
                          custom_bl_sett_addr=None,
                          no_backup=False,
                          backup_address=None,
                          app_boot_validation_type=None,
                          sd_boot_validation_type=None,
                          sd_file=None,
                          signer=None)

        self.assertEqual(nRFArch.NRF52, settings.arch)
        self.assertEqual('nRF52', settings.arch_str)
        self.assertEqual(0x0007F000, settings.bl_sett_addr)
        self.assertEqual(0xEAA3288E, settings.crc)
        self.assertEqual(0x00000001, settings.bl_sett_ver)
        self.assertEqual(0x00000001, settings.app_ver)
        self.assertEqual(0x00000001, settings.bl_ver)
        self.assertEqual(0x00000000, settings.bank_layout)
        self.assertEqual(0x00000000, settings.bank_current)
        self.assertEqual(0x00000000, settings.app_sz)
        self.assertEqual(0x00000000, settings.app_crc)
        self.assertEqual(0x00000000, settings.bank0_bank_code)

    def test_generate_with_application_file(self):
        settings = BLDFUSettings()
        settings.generate(arch='NRF52',
                          app_file='firmwares/s132_nrf52_mini.hex',
                          app_ver=1,
                          bl_ver=1,
                          bl_sett_ver=1,
                          custom_bl_sett_addr=None,
                          no_backup=False,
                          backup_address=None,
                          app_boot_validation_type=None,
                          sd_boot_validation_type=None,
                          sd_file=None,
                          signer=None)

        self.assertEqual(nRFArch.NRF52, settings.arch)
        self.assertEqual('nRF52', settings.arch_str)
        self.assertEqual(0x0007F000, settings.bl_sett_addr)
        self.assertEqual(0xEB5917FB, settings.crc)
        self.assertEqual(0x00000001, settings.bl_sett_ver)
        self.assertEqual(0x00000001, settings.app_ver)
        self.assertEqual(0x00000001, settings.bl_ver)
        self.assertEqual(0x00000000, settings.bank_layout)
        self.assertEqual(0x00000000, settings.bank_current)
        self.assertEqual(0x00003000, settings.app_sz)
        self.assertEqual(0x5F045729, settings.app_crc)
        self.assertEqual(0x00000001, settings.bank0_bank_code)

    def test_generate_with_custom_start_address(self):
        settings = BLDFUSettings()
        settings.generate(arch='NRF52',
                          app_file=None,
                          app_ver=1,
                          bl_ver=1,
                          bl_sett_ver=1,
                          custom_bl_sett_addr=0x0006F000,
                          no_backup=False,
                          backup_address=None,
                          app_boot_validation_type=None,
                          sd_boot_validation_type=None,
                          sd_file=None,
                          signer=None)

        self.assertEqual(settings.bl_sett_addr, 0x0006F000)

    def test_generate_with_backup_page_check_size(self):
        settings_raw = BLDFUSettings()
        settings_raw.generate(arch='NRF52',
                             app_file=None,
                             app_ver=1,
                             bl_ver=1,
                             bl_sett_ver=1,
                             custom_bl_sett_addr=None,
                             no_backup=True,
                             backup_address=None,
                             app_boot_validation_type=None,
                             sd_boot_validation_type=None,
                             sd_file=None,
                             signer=None)

        settings = BLDFUSettings()
        settings.generate(arch='NRF52',
                          app_file=None,
                          app_ver=1,
                          bl_ver=1,
                          bl_sett_ver=1,
                          custom_bl_sett_addr=None,
                          no_backup=False,
                          backup_address=None,
                          app_boot_validation_type=None,
                          sd_boot_validation_type=None,
                          sd_file=None,
                          signer=None)

        self.assertEqual(len(list(settings.ihex.todict().keys())), len(list(settings_raw.ihex.todict().keys())) * 2)

    def test_generate_with_backup_page_check_values(self):
        settings = BLDFUSettings()
        settings.generate(arch='NRF52',
                          app_file=None,
                          app_ver=1,
                          bl_ver=1,
                          bl_sett_ver=1,
                          custom_bl_sett_addr=None,
                          no_backup=False,
                          backup_address=None,
                          app_boot_validation_type=None,
                          sd_boot_validation_type=None,
                          sd_file=None,
                          signer=None)

        backup_dict = {(k + settings.bl_sett_backup_offset): v for k, v in list(settings.ihex.todict().items()) if k < settings.bl_sett_addr}
        settings_dict = {k: v for k, v in list(settings.ihex.todict().items()) if k >= settings.bl_sett_addr}
        self.assertEqual(backup_dict, settings_dict)

    def test_generate_with_backup_page_custom_address(self):
        settings = BLDFUSettings()
        settings.generate(arch='NRF52',
                          app_file=None,
                          app_ver=1,
                          bl_ver=1,
                          bl_sett_ver=1,
                          custom_bl_sett_addr=None,
                          no_backup=False,
                          backup_address=0x0006F000,
                          app_boot_validation_type=None,
                          sd_boot_validation_type=None,
                          sd_file=None,
                          signer=None)

        self.assertEqual(settings.backup_address, 0x0006F000)
        self.assertTrue(0x0006F000 in list(settings.ihex.todict().keys()))

    def test_generate_with_backup_page_default_address(self):
        settings = BLDFUSettings()
        settings.generate(arch='NRF52',
                          app_file=None,
                          app_ver=1,
                          bl_ver=1,
                          bl_sett_ver=1,
                          custom_bl_sett_addr=0x0006F000,
                          no_backup=False,
                          backup_address=None,
                          app_boot_validation_type=None,
                          sd_boot_validation_type=None,
                          sd_file=None,
                          signer=None)

        self.assertEqual(settings.backup_address, (0x0006F000 - settings.bl_sett_backup_offset))
        self.assertTrue((0x0006F000 - settings.bl_sett_backup_offset) in list(settings.ihex.todict().keys()))

class TestBLDFUSettingsV2(unittest.TestCase):
    def setUp(self):
        script_abspath = os.path.abspath(__file__)
        script_dirname = os.path.dirname(script_abspath)
        os.chdir(script_dirname)

    def test_fromhexfile(self):
        settings = BLDFUSettings()
        settings.fromhexfile('firmwares/bl_settings_v2_nrf52.hex')

        self.assertEqual(nRFArch.NRF52, settings.arch)
        self.assertEqual('nRF52', settings.arch_str)
        self.assertEqual(0x0007F000, settings.bl_sett_addr)
        self.assertEqual(0x2914E4A9, settings.crc)
        self.assertEqual(0x00000002, settings.bl_sett_ver)
        self.assertEqual(0x00000001, settings.app_ver)
        self.assertEqual(0x00000001, settings.bl_ver)
        self.assertEqual(0x00000000, settings.bank_layout)
        self.assertEqual(0x00000000, settings.bank_current)
        self.assertEqual(0x000148B4, settings.app_sz)
        self.assertEqual(0xF272EEBF, settings.app_crc)
        self.assertEqual(0x00000001, settings.bank0_bank_code)
        self.assertEqual(0x00024150, settings.sd_sz)
        self.assertEqual(0x467B5555, settings.boot_validation_crc)
        self.assertEqual(0x01,       settings.sd_boot_validation_type)
        #self.assertEqual(0x5B00BDCE, settings.sd_boot_validation_bytes)
        self.assertEqual(0x01,       settings.app_boot_validation_type)
        #self.assertEqual(0xF272EEBF, settings.app_boot_validation_bytes)

    def test_generate_without_application_or_sd_file(self):
        settings = BLDFUSettings()
        settings.generate(arch='NRF52',
                          app_file=None,
                          app_ver=1,
                          bl_ver=1,
                          bl_sett_ver=2,
                          custom_bl_sett_addr=None,
                          no_backup=False,
                          backup_address=None,
                          app_boot_validation_type=None,
                          sd_boot_validation_type=None,
                          sd_file=None,
                          signer=None)

        self.assertEqual(nRFArch.NRF52, settings.arch)
        self.assertEqual('nRF52', settings.arch_str)
        self.assertEqual(0x0007F000, settings.bl_sett_addr)
        self.assertEqual(0x12ECF0A5, settings.crc)
        self.assertEqual(0x00000002, settings.bl_sett_ver)
        self.assertEqual(0x00000001, settings.app_ver)
        self.assertEqual(0x00000001, settings.bl_ver)
        self.assertEqual(0x00000000, settings.bank_layout)
        self.assertEqual(0x00000000, settings.bank_current)
        self.assertEqual(0x00000000, settings.app_sz)
        self.assertEqual(0x00000000, settings.app_crc)
        self.assertEqual(0x00000000, settings.bank0_bank_code)
        self.assertEqual(0x00000000, settings.sd_sz)
        self.assertEqual(0xACDA1BA2, settings.boot_validation_crc)
        self.assertEqual(0x00, settings.sd_boot_validation_type)
        # self.assertEqual(0x5B00BDCE, settings.sd_boot_validation_bytes)
        self.assertEqual(0x00, settings.app_boot_validation_type)
        # self.assertEqual(0xF272EEBF, settings.app_boot_validation_bytes)

    def test_generate_with_application_and_sd_file(self):
        settings = BLDFUSettings()
        settings.generate(arch='NRF52',
                          app_file='firmwares/s132_nrf52_mini.hex',
                          app_ver=1,
                          bl_ver=1,
                          bl_sett_ver=2,
                          custom_bl_sett_addr=None,
                          no_backup=False,
                          backup_address=None,
                          app_boot_validation_type=None,
                          sd_boot_validation_type=None,
                          sd_file='firmwares/s132_nrf52_mini.hex',
                          signer=None)

        self.assertEqual(nRFArch.NRF52, settings.arch)
        self.assertEqual('nRF52', settings.arch_str)
        self.assertEqual(0x0007F000, settings.bl_sett_addr)
        self.assertEqual(0x47CD3EEA, settings.crc)
        self.assertEqual(0x00000002, settings.bl_sett_ver)
        self.assertEqual(0x00000001, settings.app_ver)
        self.assertEqual(0x00000001, settings.bl_ver)
        self.assertEqual(0x00000000, settings.bank_layout)
        self.assertEqual(0x00000000, settings.bank_current)
        self.assertEqual(0x00003000, settings.app_sz)
        self.assertEqual(0x5F045729, settings.app_crc)
        self.assertEqual(0x00000001, settings.bank0_bank_code)
        self.assertEqual(0x00003000, settings.sd_sz)
        self.assertEqual(0xACDA1BA2, settings.boot_validation_crc)
        self.assertEqual(0x00, settings.sd_boot_validation_type)
        # self.assertEqual(0x5B00BDCE, settings.sd_boot_validation_bytes)
        self.assertEqual(0x00, settings.app_boot_validation_type)
        # self.assertEqual(0xF272EEBF, settings.app_boot_validation_bytes)

    def test_generate_with_custom_start_address(self):
        settings = BLDFUSettings()
        settings.generate(arch='NRF52',
                          app_file=None,
                          app_ver=1,
                          bl_ver=1,
                          bl_sett_ver=2,
                          custom_bl_sett_addr=0x0006F000,
                          no_backup=False,
                          backup_address=None,
                          app_boot_validation_type=None,
                          sd_boot_validation_type=None,
                          sd_file=None,
                          signer=None)

        self.assertEqual(settings.bl_sett_addr, 0x0006F000)

    def test_generate_with_backup_page_check_size(self):
        settings_raw = BLDFUSettings()
        settings_raw.generate(arch='NRF52',
                             app_file=None,
                             app_ver=1,
                             bl_ver=1,
                             bl_sett_ver=2,
                             custom_bl_sett_addr=None,
                             no_backup=True,
                             backup_address=None,
                             app_boot_validation_type=None,
                             sd_boot_validation_type=None,
                             sd_file=None,
                             signer=None)

        settings = BLDFUSettings()
        settings.generate(arch='NRF52',
                          app_file=None,
                          app_ver=1,
                          bl_ver=1,
                          bl_sett_ver=2,
                          custom_bl_sett_addr=None,
                          no_backup=False,
                          backup_address=None,
                          app_boot_validation_type=None,
                          sd_boot_validation_type=None,
                          sd_file=None,
                          signer=None)

        self.assertEqual(len(list(settings.ihex.todict().keys())), len(list(settings_raw.ihex.todict().keys())) * 2)

    def test_generate_with_backup_page_check_values(self):
        settings = BLDFUSettings()
        settings.generate(arch='NRF52',
                          app_file=None,
                          app_ver=1,
                          bl_ver=1,
                          bl_sett_ver=2,
                          custom_bl_sett_addr=None,
                          no_backup=False,
                          backup_address=None,
                          app_boot_validation_type=None,
                          sd_boot_validation_type=None,
                          sd_file=None,
                          signer=None)

        backup_dict = {(k + settings.bl_sett_backup_offset): v for k, v in list(settings.ihex.todict().items()) if k < settings.bl_sett_addr}
        settings_dict = {k: v for k, v in list(settings.ihex.todict().items()) if k >= settings.bl_sett_addr}
        self.assertEqual(backup_dict, settings_dict)

    def test_generate_with_backup_page_custom_address(self):
        settings = BLDFUSettings()
        settings.generate(arch='NRF52',
                          app_file=None,
                          app_ver=1,
                          bl_ver=1,
                          bl_sett_ver=2,
                          custom_bl_sett_addr=None,
                          no_backup=False,
                          backup_address=0x0006F000,
                          app_boot_validation_type=None,
                          sd_boot_validation_type=None,
                          sd_file=None,
                          signer=None)

        self.assertEqual(settings.backup_address, 0x0006F000)
        self.assertTrue(0x0006F000 in list(settings.ihex.todict().keys()))

    def test_generate_with_backup_page_default_address(self):
        settings = BLDFUSettings()
        settings.generate(arch='NRF52',
                          app_file=None,
                          app_ver=1,
                          bl_ver=1,
                          bl_sett_ver=2,
                          custom_bl_sett_addr=0x0006F000,
                          no_backup=False,
                          backup_address=None,
                          app_boot_validation_type=None,
                          sd_boot_validation_type=None,
                          sd_file=None,
                          signer=None)

        self.assertEqual(settings.backup_address, (0x0006F000 - settings.bl_sett_backup_offset))
        self.assertTrue((0x0006F000 - settings.bl_sett_backup_offset) in list(settings.ihex.todict().keys()))

    def test_generate_with_app_boot_validation_crc(self):
        settings = BLDFUSettings()
        settings.generate(arch='NRF52',
                          app_file='firmwares/s132_nrf52_mini.hex',
                          app_ver=1,
                          bl_ver=1,
                          bl_sett_ver=2,
                          custom_bl_sett_addr=None,
                          no_backup=True,
                          backup_address=None,
                          app_boot_validation_type='VALIDATE_GENERATED_CRC',
                          sd_boot_validation_type=None,
                          sd_file=None,
                          signer=None)

        self.assertEqual(0x1316CFD0, settings.crc)
        self.assertEqual(0x49A0F45A, settings.boot_validation_crc)
        self.assertEqual(0x01,       settings.app_boot_validation_type)
        self.assertEqual(0x5F045729, struct.unpack('<I', settings.app_boot_validation_bytes)[0])

    def test_generate_with_app_boot_validation_sha256(self):
        settings = BLDFUSettings()
        settings.generate(arch='NRF52',
                          app_file='firmwares/s132_nrf52_mini.hex',
                          app_ver=1,
                          bl_ver=1,
                          bl_sett_ver=2,
                          custom_bl_sett_addr=None,
                          no_backup=True,
                          backup_address=None,
                          app_boot_validation_type='VALIDATE_GENERATED_SHA256',
                          sd_boot_validation_type=None,
                          sd_file=None,
                          signer=None)

        self.assertEqual(0x1316CFD0, settings.crc)
        self.assertEqual(0xF78E451E, settings.boot_validation_crc)
        self.assertEqual(0x02, settings.app_boot_validation_type)
        self.assertEqual(bytes.fromhex('036F52C9EBB53819D6E2B6FB57803823E864783B04D7331B46C0B5897CA9F1C7'),
                         settings.app_boot_validation_bytes)

    def test_generate_with_app_boot_validation_ecdsa(self):
        settings = BLDFUSettings()

        signer = Signing()
        signer.load_key('key.pem')

        settings.generate(arch='NRF52',
                          app_file='firmwares/s132_nrf52_mini.hex',
                          app_ver=1,
                          bl_ver=1,
                          bl_sett_ver=2,
                          custom_bl_sett_addr=None,
                          no_backup=True,
                          backup_address=None,
                          app_boot_validation_type='VALIDATE_ECDSA_P256_SHA256',
                          sd_boot_validation_type=None,
                          sd_file=None,
                          signer=signer)

        # Since ECDSA contains a random component the signature will be different every time
        # it is generated. Therefore only overall structure of the boot validation will be checked.
        self.assertEqual(0x03, settings.app_boot_validation_type)
        self.assertEqual(64, len(settings.app_boot_validation_bytes))

    def test_generate_with_sd_boot_validation_crc(self):
        settings = BLDFUSettings()
        settings.generate(arch='NRF52',
                          app_file=None,
                          app_ver=1,
                          bl_ver=1,
                          bl_sett_ver=2,
                          custom_bl_sett_addr=None,
                          no_backup=True,
                          backup_address=None,
                          app_boot_validation_type=None,
                          sd_boot_validation_type='VALIDATE_GENERATED_CRC',
                          sd_file='firmwares/s132_nrf52_mini.hex',
                          signer=None)

        self.assertEqual(0x4637019F, settings.crc)
        self.assertEqual(0xCB5F90FB, settings.boot_validation_crc)
        self.assertEqual(0x01,       settings.sd_boot_validation_type)
        self.assertEqual(0x5F045729, struct.unpack('<I', settings.sd_boot_validation_bytes)[0])

    def test_generate_with_sd_boot_validation_sha256(self):
        settings = BLDFUSettings()
        settings.generate(arch='NRF52',
                          app_file=None,
                          app_ver=1,
                          bl_ver=1,
                          bl_sett_ver=2,
                          custom_bl_sett_addr=None,
                          no_backup=True,
                          backup_address=None,
                          app_boot_validation_type=None,
                          sd_boot_validation_type='VALIDATE_GENERATED_SHA256',
                          sd_file='firmwares/s132_nrf52_mini.hex',
                          signer=None)

        self.assertEqual(0x4637019F, settings.crc)
        self.assertEqual(0x9C761426, settings.boot_validation_crc)
        self.assertEqual(0x02, settings.sd_boot_validation_type)
        self.assertEqual(bytes.fromhex('036F52C9EBB53819D6E2B6FB57803823E864783B04D7331B46C0B5897CA9F1C7'),
                         settings.sd_boot_validation_bytes)

    def test_generate_with_sd_boot_validation_ecdsa(self):
        settings = BLDFUSettings()

        signer = Signing()
        signer.load_key('key.pem')

        settings.generate(arch='NRF52',
                          app_file=None,
                          app_ver=1,
                          bl_ver=1,
                          bl_sett_ver=2,
                          custom_bl_sett_addr=None,
                          no_backup=True,
                          backup_address=None,
                          app_boot_validation_type=None,
                          sd_boot_validation_type='VALIDATE_ECDSA_P256_SHA256',
                          sd_file='firmwares/s132_nrf52_mini.hex',
                          signer=signer)

        # Since ECDSA contains a random component the signature will be different every time
        # it is generated. Therefore only overall structure of the boot validation will be checked.
        self.assertEqual(0x03, settings.sd_boot_validation_type)
        self.assertEqual(64, len(settings.sd_boot_validation_bytes))

if __name__ == '__main__':
    unittest.main()



#
# Copyright (c) 2016 Nordic Semiconductor ASA
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#   1. Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright notice, this
#   list of conditions and the following disclaimer in the documentation and/or
#   other materials provided with the distribution.
#
#   3. Neither the name of Nordic Semiconductor ASA nor the names of other
#   contributors to this software may be used to endorse or promote products
#   derived from this software without specific prior written permission.
#
#   4. This software must only be used in or with a processor manufactured by Nordic
#   Semiconductor ASA, or in or with a processor manufactured by a third party that
#   is used in combination with a processor manufactured by Nordic Semiconductor.
#
#   5. Any software provided in binary or object form under this license must not be
#   reverse engineered, decompiled, modified and/or disassembled.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import json
import os
import tempfile
import unittest
from zipfile import ZipFile
import shutil

from nordicsemi.dfu.package import Package
from nordicsemi.dfu.signing import Signing


class TestPackage(unittest.TestCase):
    def setUp(self):
        self.work_directory = tempfile.mkdtemp(prefix="nrf_dfu_tests_")

    def tearDown(self):
        shutil.rmtree(self.work_directory, ignore_errors=True)

    def test_generate_package_application(self):
        signer = Signing()
        signer.load_key('key.pem')

        self.p = Package(app_version=100,
            sd_req=[0x1000, 0xfffe],
            app_fw="firmwares/bar.hex",
            signer=signer
        )

        pkg_name = "mypackage.zip"

        self.p.generate_package(pkg_name, preserve_work_dir=False)
        expected_zip_content = ["manifest.json", "bar.bin", "bar.dat"]

        with ZipFile(pkg_name, 'r') as pkg:
            infolist = pkg.infolist()

            for file_information in infolist:
                self.assertTrue(file_information.filename in expected_zip_content)
                self.assertGreater(file_information.file_size, 0)

            # Extract all and load json document to see if it is correct regarding to paths
            pkg.extractall(self.work_directory)

            with open(os.path.join(self.work_directory, 'manifest.json'), 'r') as f:
                _json = json.load(f)
                self.assertEqual('bar.bin', _json['manifest']['application']['bin_file'])
                self.assertEqual('bar.dat', _json['manifest']['application']['dat_file'])
                self.assertTrue('softdevice' not in _json['manifest'])
                self.assertTrue('softdevice_bootloader' not in _json['manifest'])
                self.assertTrue('bootloader' not in _json['manifest'])

    def test_generate_package_sd_bl(self):
        signer = Signing()
        signer.load_key('key.pem')

        self.p = Package(app_version=100,
                         sd_req=[0x1000, 0xfffe],
                         softdevice_fw="firmwares/foo.hex",
                         bootloader_fw="firmwares/bar.hex",
                         signer=signer)


        pkg_name = "mypackage.zip"

        self.p.generate_package(pkg_name, preserve_work_dir=False)

        expected_zip_content = ["manifest.json", "sd_bl.bin", "sd_bl.dat"]

        with ZipFile(pkg_name, 'r') as pkg:
            infolist = pkg.infolist()

            for file_information in infolist:
                self.assertTrue(file_information.filename in expected_zip_content)
                self.assertGreater(file_information.file_size, 0)

            # Extract all and load json document to see if it is correct regarding to paths
            pkg.extractall(self.work_directory)

            with open(os.path.join(self.work_directory, 'manifest.json'), 'r') as f:
                _json = json.load(f)
                self.assertEqual('sd_bl.bin', _json['manifest']['softdevice_bootloader']['bin_file'])
                self.assertEqual('sd_bl.dat', _json['manifest']['softdevice_bootloader']['dat_file'])

    def test_unpack_package_a(self):
        signer = Signing()
        signer.load_key('key.pem')

        self.p = Package(app_version=100,
                         sd_req=[0x1000, 0xffff],
                         softdevice_fw="firmwares/bar.hex",
                         signer=signer)
        pkg_name = os.path.join(self.work_directory, "mypackage.zip")
        self.p.generate_package(pkg_name, preserve_work_dir=False)

        unpacked_dir = os.path.join(self.work_directory, "unpacked")
        manifest = self.p.unpack_package(os.path.join(self.work_directory, pkg_name), unpacked_dir)
        self.assertIsNotNone(manifest)
        self.assertEqual('bar.bin', manifest.softdevice.bin_file)
#         self.assertEqual(0, manifest.softdevice.init_packet_data.ext_packet_id)
#         self.assertIsNotNone(manifest.softdevice.init_packet_data.firmware_crc16)

    def test_unpack_package_b(self):
        signer = Signing()
        signer.load_key('key.pem')

        self.p = Package(app_version=100,
                         sd_req=[0x1000, 0xffff],
                         softdevice_fw="firmwares/bar.hex",
                         signer=signer)
        pkg_name = os.path.join(self.work_directory, "mypackage.zip")
        self.p.generate_package(pkg_name, preserve_work_dir=False)

        unpacked_dir = os.path.join(self.work_directory, "unpacked")
        manifest = self.p.unpack_package(os.path.join(self.work_directory, pkg_name), unpacked_dir)
        self.assertIsNotNone(manifest)
        self.assertEqual('bar.bin', manifest.softdevice.bin_file)

    def test_unpack_package_c(self):
        signer = Signing()
        signer.load_key('key.pem')

        self.p = Package(app_version=100,
                         sd_req=[0x1000, 0xffff],
                         softdevice_fw="firmwares/bar.hex",
                         signer=signer)
        pkg_name = os.path.join(self.work_directory, "mypackage.zip")
        self.p.generate_package(pkg_name, preserve_work_dir=False)

        unpacked_dir = os.path.join(self.work_directory, "unpacked")
        manifest = self.p.unpack_package(os.path.join(self.work_directory, pkg_name), unpacked_dir)
        self.assertIsNotNone(manifest)
        self.assertEqual('bar.bin', manifest.softdevice.bin_file)


if __name__ == '__main__':
    unittest.main()
